import os
import re
import warnings
from pathlib import Path

import pytest
from astropy.coordinates import SkyCoord
from requests import ConnectionError, ReadTimeout

from stellarphot.io.tess import (
    TOI,
    TessSubmission,
    TessTargetFile,
    tess_photometry_setup,
)

GOOD_HEADER = {
    "date-obs": "2022-06-04T05:44:28.010",
    "filter": "ip",
    "object": "TIC-237205154",
}

GOOD_HEADER_WITH_PLANET = {
    "date-obs": "2022-06-04T05:44:28.010",
    "filter": "ip",
    "object": "TIC-237205154.01",
}

BAD_HEADER = {}


def test_good_header_sucess():
    tsub = TessSubmission.from_header(GOOD_HEADER)
    assert tsub.utc_start == "20220604"
    assert tsub.tic_id == 237205154
    assert tsub.planet_number == 0
    assert tsub.telescope_code == ""


def test_good_header_planet_sucess():
    tsub = TessSubmission.from_header(GOOD_HEADER_WITH_PLANET)
    assert tsub.utc_start == "20220604"
    assert tsub.tic_id == 237205154
    assert tsub.planet_number == 1
    assert tsub.telescope_code == ""


def test_bad_header_fails():
    with pytest.raises(ValueError) as e:
        TessSubmission.from_header(BAD_HEADER)
    match_str = ".*UTC date of first image.*filter/passband.*TIC ID number.*"
    assert re.search(match_str, str(e))


def test_base_name():
    tsub = TessSubmission.from_header(GOOD_HEADER_WITH_PLANET, telescope_code="ABS")
    assert tsub.base_name == "TIC237205154-01_20220604_ABS_ip"


def test_seeing_profile():
    tsub = TessSubmission.from_header(GOOD_HEADER_WITH_PLANET, telescope_code="ABS")
    assert tsub.seeing_profile == "TIC237205154-01_20220604_ABS_ip_seeing-profile.png"


def test_valid_method():
    tsub = TessSubmission.from_header(GOOD_HEADER)

    # Missing telescope code, amopng other things...
    assert not tsub._valid()

    # Fix the code, should still not be valid because of planet number
    tsub.telescope_code = "ABC"
    assert not tsub._valid()

    # Set planet number and should be valid
    tsub.planet_number = 1
    assert tsub._valid()

    # Set invalid TIC number
    tsub.tic_id = 10_000_000_000
    assert not tsub._valid()


@pytest.mark.remote_data
def test_target_file():
    # Getting the target information failed on windows, so the
    # first point of this test is to simply succeed in creating the
    # object

    tic_742648307 = SkyCoord(ra=104.733225, dec=49.968739, unit="degree")
    server_down = False

    try:
        tess_target = TessTargetFile(tic_742648307, magnitude=12, depth=10)
    except (ConnectionError, ReadTimeout):
        server_down = True
        tess_target = None  # Assure tess_target is defined so that we can delete it
    except ValueError:
        # The server is technically back but producing garbage....
        server_down = True
        tess_target = None  # Assure tess_target is defined so that we can delete it
    else:
        # Check that the first thing in the list is the tick object
        check_coords = SkyCoord(
            ra=tess_target.table["RA"][0],
            dec=tess_target.table["Dec"][0],
            unit=("hour", "degree"),
        )
        assert tic_742648307.separation(check_coords).arcsecond < 1
    finally:
        # Make sure that regardless of whether the server is down we delete
        # the temporary file that gets created.

        # On windows the temporary file cannot be deleted because if it is
        # then the file immediately vanished. That means we get a warning
        # about having an open file handle. We can ignore that warning
        # because we don't care about it here.

        # The warning doesn't get generated until the last reference to
        # the object is deleted, so we need to do that here instead of
        # letting it happen at the end of the test.
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="unclosed file", category=ResourceWarning
            )
            try:
                del tess_target
            except UnboundLocalError:
                pass

        if server_down:
            # The server at University of Louisville is down sometimes, so
            # xfail this test when that happens
            pytest.xfail("TESS/gaia Server down")


class TestTOI:
    @pytest.fixture
    def sample_toi(self):
        toi_dict = {
            "tic_id": 236158940,
            "coord": {
                "ra": "313d25m10.33459171s",
                "dec": "34d21m05.92981069s",
                "representation_type": "spherical",
                "frame": "icrs",
            },
            "depth_ppt": 3.31,
            "depth_error_ppt": 0.13131700000000002,
            "duration": "2.885 h",
            "duration_error": "0.286 h",
            "epoch": {
                "jd1": 2459817.0,
                "jd2": 0.28766999999061227,
                "format": "jd",
                "scale": "tdb",
                "precision": 3,
                "in_subfmt": "*",
                "out_subfmt": "*",
            },
            "epoch_error": "0.0018336 d",
            "period": "2.677573 d",
            "period_error": "1.8e-05 d",
            "tess_mag": 11.3283,
            "tess_mag_error": 0.015,
        }
        return TOI.model_validate(toi_dict)

    @pytest.mark.remote_data
    def test_from_tic_id(self, tess_tic_expected_values):
        # Nothing special about the TIC ID chosen here. It is one we happened
        # to be looking at when writing this test.
        tic_id = tess_tic_expected_values["tic_id"]
        toi_info = TOI.from_tic_id(tic_id)
        assert toi_info.tic_id == tic_id

        # Test the coordinate, but not other properties because those may change
        # over time as the planet candidate is better refined. The coordinate
        # should always be the same.
        expected_coord = tess_tic_expected_values["expected_coords"]
        assert toi_info.coord.separation(expected_coord).arcsecond < 1

        # Try round-tripping through json
        # This is a regression test for #427
        json_str = toi_info.model_dump_json()
        new_toi = TOI.model_validate_json(json_str)

        assert toi_info.coord.separation(new_toi.coord).arcsecond < 0.01


class TestTessPhotometrySetup:
    # This auto-used fixture changes the working directory to the temporary directory
    # and then changes back to the original directory after the test is done.
    @pytest.fixture(autouse=True)
    def change_to_tmp_dir(self, tmp_path):
        original_dir = os.getcwd()
        os.chdir(tmp_path)
        # Yielding here is important. It means that when the test is done, the remainder
        # of the function will be executed. This is important because the test is run in
        # a temporary directory and we want to change back to the original directory
        # when the test is done.
        yield
        os.chdir(original_dir)

    def test_creation_invalid_input_raises_error(self):
        with pytest.raises(
            ValueError, match="Must provide either TIC ID or TOI object"
        ):
            tess_photometry_setup()

    @pytest.mark.remote_data
    @pytest.mark.parametrize("creation_method", ["tic_id", "toi_object"])
    def test_creation(self, tess_tic_expected_values, creation_method):
        # Check that we can create the necessary files for TESS photometry
        # from a TIC ID.
        tic_id = tess_tic_expected_values["tic_id"]
        if creation_method == "tic_id":
            tess_photometry_setup(tic_id=tic_id)
        else:
            toi_info = TOI.from_tic_id(tic_id)
            tess_photometry_setup(TOI_object=toi_info)

        p_info = Path(f"TIC-{tic_id}-info.json")
        assert p_info.exists()

        p_source_list = Path(f"TIC-{tic_id}-source-list-input.ecsv")
        assert p_source_list.exists()

    @pytest.mark.remote_data
    def test_creation_with_overwrite(self, tess_tic_expected_values):
        # Check to see that an error is raised if the files already exist
        # and the overwrite flag is not set.
        tic_id = tess_tic_expected_values["tic_id"]
        tess_photometry_setup(tic_id=tic_id)

        # Try re-running with both files present, where we hit the error from
        # the source list file first
        with pytest.raises(FileExistsError):
            tess_photometry_setup(tic_id=tic_id)

        # Remove the source list table and try again, this time hitting the
        # error from the info file
        Path(f"TIC-{tic_id}-source-list-input.ecsv").unlink()
        with pytest.raises(FileExistsError):
            tess_photometry_setup(tic_id=tic_id)



import os
from pathlib import Path

import pytest
from astropy.coordinates import SkyCoord
from astropy.io import fits

from stellarphot import SourceListData
from stellarphot.io import TOI
from stellarphot.settings.custom_widgets import TessPhotometrySetup


class TestTessPhotometrySetup:
    # This auto-used fixture changes the working directory to the temporary directory
    # and then changes back to the original directory after the test is done.
    @pytest.fixture(autouse=True)
    def change_to_tmp_dir(self, tmp_path):
        original_dir = os.getcwd()
        os.chdir(tmp_path)
        # Yielding here is important. It means that when the test is done, the remainder
        # of the function will be executed. This is important because the test is run in
        # a temporary directory and we want to change back to the original directory
        # when the test is done.
        yield
        os.chdir(original_dir)

    def test_creation(self):
        widget = TessPhotometrySetup()
        # Initial TIC ID should be 0
        assert widget.tic_id == 0

        # All done display should be hidden initially
        assert widget.all_done.layout.display == "none"

        # Selection method should be set to option 0
        assert widget.drop.value == 0

    def test_choose_tic_id_entry_method(self):
        widget = TessPhotometrySetup()
        # Choose the TIC ID entry method
        widget.drop.value = 0

        # The TIC ID entry should be visible
        assert widget.tic_id_entry.layout.display == "flex"
        # The FITS opener should be hidden
        assert widget.fits_opener.file_chooser.layout.display == "none"

        # now choose the FITS opener method
        widget.drop.value = 1

        # The TIC ID entry should be hidden
        assert widget.tic_id_entry.layout.display == "none"
        # The FITS opener should be visible
        assert widget.fits_opener.file_chooser.layout.display == "flex"

    def test_set_tic_id_via_text_box(self, tess_tic_expected_values):
        widget = TessPhotometrySetup()
        # This needs to be a real TIC ID....
        widget.tic_id_entry.value = tess_tic_expected_values["tic_id"]
        assert widget.tic_id == tess_tic_expected_values["tic_id"]

        # The confirmation dialog should be visible now
        assert widget.confirm.layout.display == "flex"

    def test_get_tic_id_via_file(self, tess_tic_expected_values):
        widget = TessPhotometrySetup()

        # Make a little FITS file with an object keyword
        hdu = fits.PrimaryHDU()
        hdu.header["object"] = f"TIC {tess_tic_expected_values['tic_id']}"

        fits_file = Path("test.fits")

        hdu.writeto(fits_file)

        widget.fits_opener.file_chooser.value = str(fits_file)

        assert widget.tic_id == tess_tic_expected_values["tic_id"]

        # The confirmation dialog should be visible now
        assert widget.confirm.layout.display == "flex"

    def test_click_no_confirm(self, tess_tic_expected_values):
        # Test that clicking "No" on the confirmation dialog hides the dialog
        # and does not display the "All Done" display
        widget = TessPhotometrySetup()
        widget.tic_id_entry.value = tess_tic_expected_values["tic_id"]

        # Confirmation box is displayed, click no
        widget.confirm._no.click()

        # The confirmation dialog should be hidden now
        assert widget.confirm.layout.display == "none"

        # The all done display should still be hidden
        assert widget.all_done.layout.display == "none"

    @pytest.mark.remote_data
    def test_click_yes_confirm(self, tess_tic_expected_values):
        # Test that clicking "Yes" on the confirmation dialog hides the dialog
        # and displays the "All Done" display and creates the files we expect.

        widget = TessPhotometrySetup()
        widget.tic_id_entry.value = tess_tic_expected_values["tic_id"]

        # Confirmation box is displayed, click yes
        widget.confirm._yes.click()

        # The confirmation dialog should be hidden now
        assert widget.confirm.layout.display == "none"

        # The all done display should be visible
        assert widget.all_done.layout.display == "flex"

        # There should be exactly one json file from which we can create a TOI object
        TOI_json_file = list(Path(".").glob("*.json"))
        assert len(TOI_json_file) == 1
        TOI_json_file = TOI_json_file[0]
        assert str(tess_tic_expected_values["tic_id"]) in TOI_json_file.name

        # Make a TOI object from the JSON
        print(TOI_json_file.read_text())
        toi = TOI.model_validate_json(TOI_json_file.read_text())
        # Check that the TIC ID is correct
        assert toi.tic_id == tess_tic_expected_values["tic_id"]

        # There should also be one source list file
        source_list_file = list(Path(".").glob("*.ecsv"))
        assert len(source_list_file) == 1
        source_list_file = source_list_file[0]

        # This should not raise any errors
        sld = SourceListData.read(source_list_file)

        # THe first entry should be the exoplanet candidate
        assert toi.coord.separation(SkyCoord(sld["ra"][0], sld["dec"][0])).arcsecond < 1



import numpy as np
from astropy import units as u
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.modeling.models import Const2D, Gaussian2D
from astropy.nddata import CCDData
from astropy.nddata.utils import Cutout2D
from astropy.stats import gaussian_sigma_to_fwhm, sigma_clipped_stats
from astropy.table import Table
from photutils.detection import DAOStarFinder
from photutils.morphology import data_properties

from stellarphot.core import SourceListData

__all__ = ["source_detection", "compute_fwhm"]


def _fit_2dgaussian(data):
    """
    Fit a 2D Gaussian to data.

    Written as a replacement for functionality that was removed from
    photutils.

    This function will be kept private so we don't have to support it.

    Copy/pasted from
    https://github.com/astropy/photutils/pull/1064/files#diff-9e64908ff7ac552845b4831870a749f397d73c681d218267bd9087c7757e6dd4R285

    Parameters
    ----------
    data : array-like
        The 2D array of data to fit.

    Returns
    -------
    gfit : `astropy.modeling.Model`
        The best-fit 2D Gaussian model.
    """
    mask = ~np.isfinite(data)

    # If there are non-finite pixels they need to be masked out
    # or many of the returned properties will be NaN.
    props = data_properties(data - np.min(data[~mask]), mask=mask)

    init_const = 0.0  # subtracted data minimum above
    # ptp = peak-to-peak, i.e. max - min, need to also exclude non-finite
    # values here.
    init_amplitude = np.ptp(data[~mask])

    g_init = Const2D(init_const) + Gaussian2D(
        amplitude=init_amplitude,
        x_mean=props.xcentroid,
        y_mean=props.ycentroid,
        x_stddev=props.semimajor_sigma.value,
        y_stddev=props.semiminor_sigma.value,
        theta=props.orientation.value,
    )

    fitter = LevMarLSQFitter()
    y, x = np.indices(data.shape)
    # Call fitter, enabling filtering of NaN/inf values
    gfit = fitter(g_init, x, y, data, filter_non_finite=True)

    return gfit


def compute_fwhm(
    ccd,
    sources,
    fwhm_estimate=5,
    x_column="xcenter",
    y_column="ycenter",
    fit=True,
    sky_per_pix_avg=0,
):
    """
    Computes the FWHM in both x and y directions of sources in an image.

    Parameters
    ----------

    ccd : `astropy.nddata.CCDData`
        The CCD Image array.

    sources : `astropy.table.Table`
        An astropy table of the positions of sources in the image.

    fwhm_estimate : float, optional (default=5)
        The initial guess for the FWHM of the sources in the image.

    x_column : str, optional (default='xcenter')
        The name of the column in `sources` that contains the x positions
        of the sources.

    y_column : str, optional (default='ycenter')
        The name of the column in `sources` that contains the y positions
        of the sources.

    fit : bool, optional (default=True)
        If ``True``, fit a 2D Gaussian to each source to estimate its FWHM. If
        ``False``, estimate the FWHM of each source by computing the second
        moments of its light distribution using photutils.

    sky_per_pix_avg : float or array-like of float, optional (default=0)
        Sky background to subtract before centroiding.

    Returns
    -------

    fwhm_x, fwhm_y : tuple of np.array float
        The FWHM of each source in the x and y directions.

    """

    fwhm_x = []
    fwhm_y = []
    for source in sources:
        x = source[x_column]
        y = source[y_column]
        sky = sky_per_pix_avg
        # Cutout2D needs no units on the center position, so remove unit
        # if it is present.
        try:
            x = x.value
            y = y.value
            sky = sky.value
        except AttributeError:
            pass

        cutout = Cutout2D(ccd.data, (x, y), 5 * fwhm_estimate)
        if fit:
            fit = _fit_2dgaussian(cutout.data)
            fwhm_x.append(gaussian_sigma_to_fwhm * fit.x_stddev_1)
            fwhm_y.append(gaussian_sigma_to_fwhm * fit.y_stddev_1)
            # print('Still fitting!!')
        else:
            sc = data_properties(cutout.data - sky)

            # dat = np.where(cutout.data - sky > 0, cutout.data - sky, 0)
            # mom1 = _moments(dat, order=1)
            # xc = mom1[0, 1] / mom1[0, 0]
            # yc = mom1[1, 0] / mom1[0, 0]
            # moments = _moments_central(dat,
            #                            center=(xc, yc), order=2)
            # mom_scale = (moments / mom1[0, 0])
            # fwhm_xm = 2 * np.sqrt(np.log(2) * mom_scale[0, 2])
            # fwhm_ym = 2 * np.sqrt(np.log(2) * mom_scale[2, 0])

            fwhm_xm = sc.fwhm.value
            fwhm_ym = fwhm_xm
            fwhm_x.append(fwhm_xm)
            fwhm_y.append(fwhm_ym)

    return np.array(fwhm_x), np.array(fwhm_y)


def source_detection(
    ccd,
    fwhm=8,
    sigma=3.0,
    iters=5,
    threshold=10.0,
    find_fwhm=True,
    sky_per_pix_avg=0,
    padding=0,
):
    """
    Returns an SourceListData object containing the position of sources
    within the image identified using `photutils.DAOStarFinder` algorithm.
    It can also compute the FWHM of each source by fitting a 2D Gaussian
    which can be useful for choosing a useful aperture size for photometry.

    Parameters
    ----------

    ccd : `numpy.ndarray` or `astropy.nddata.CCDData`
        This is the 2-D array of the image to be analyzed.  If
        a CCDData object is passed, its data attribute will be used.

    fwhm : float, optional (default=8)
        Initial estimate of full-width half-max of stars in the image,
        in pixels.

    sigma : float, optional. (default=3.0)
        The number of standard deviations to use as the lower and
        upper clipping limit when calculating the image background.

    iters : int, optional (default=5)
        The number of iterations to perform sigma clipping

    threshold : float, optional. (default=10.0)
        The number of standard deviations above which to select sources.

    find_fwhm : bool, optional (default=True)
        If ``True``, estimate the FWHM of each source by fitting a 2D Gaussian
        to it. If ``False``, the 'fwhm_x', 'fwhm_y', and 'width' columns of
        output table will be filled with NaN values.

    sky_per_pix_avg : float or None, optional (default=None)
        Sky background to subtract before centroiding. If set to ``None``,
        it will be estimated using the mean of the sigma_clipped_stats
        of the image.

    padding : int, optional (default=0)
        Distance from the edge of the image to ignore when searching for
        sources.

    Returns
    -------

    sources: `stellardev.SourceListData`
        A table of the positions of sources in the image.  If `find_fwhm` is
        ``True``, includes columns for `fwhm_x`, `fwhm_y, and `width`.  If
        the input CCDData object has WCS information, the columns `ra` and
        `dec` are also populated, otherwise they are set to NaN.
    """
    # if image is a CCDData object, extract the data array
    if not isinstance(ccd, CCDData) and not isinstance(ccd, np.ndarray):
        raise ValueError("ccd must be a numpy array or CCDData object")

    # If user uses units for fwhm or sky_per_pix_avg, convert to value
    if isinstance(sky_per_pix_avg, u.Quantity):
        sky_per_pix_avg = sky_per_pix_avg.value
    if isinstance(fwhm, u.Quantity):
        fwhm = fwhm.value

    # Get statistics of the input image (and use them to estimate the sky background
    # if not provided).  Using clipped stats should hopefully get rid of any
    # bright stars that might be in the image, so the mean should be a good
    # estimate of the sky background.
    print(
        "source_detection: You may see a warning about invalid values in the "
        "input image.  This is expected if any pixels are saturated and can be "
        "ignored."
    )
    mean, median, std = sigma_clipped_stats(ccd, sigma=sigma, maxiters=iters)
    print(
        f"source_detection: sigma_clipped_stats mean={mean:.4f}, "
        f"median={median:.4f}, std={std:.4f}"
    )
    if sky_per_pix_avg is None:
        sky_per_pix_avg = mean
        print(f"source_detection: sky_per_pix_avg set to {sky_per_pix_avg:.4f}")

    # Identify sources applying DAOStarFinder to a "sky subtracted"
    # image.
    print(
        f"source_detection: threshold set to {threshold}* standard deviation "
        f"({std:.4f})"
    )
    print(f"source_detection: Assuming fwhm of {fwhm} for DAOStarFinder")
    # daofind should be run on background subtracted image
    # (fails, or at least returns garbage, if sky_per_pix_avg is too low)
    daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold * std)
    if isinstance(ccd, CCDData):
        sources = daofind(ccd.data - sky_per_pix_avg)
    else:
        sources = daofind(ccd - sky_per_pix_avg)

    # Identify sources near the edge of the image and remove them
    # from the source list.
    padding_smt = ""
    if padding > 0:
        src_cnt0 = len(sources)
        y_lim, x_lim = ccd.shape
        keep = (
            (sources["xcentroid"].value >= padding)
            & (sources["ycentroid"].value >= padding)
            & (sources["xcentroid"].value < x_lim - padding)
            & (sources["ycentroid"].value < y_lim - padding)
        )
        sources = sources[keep]
        padding_smt = f" (after removing {src_cnt0-len(sources)} sources near the edge)"

    src_cnt = len(sources)
    print(f"source_detection: {src_cnt} sources identified{padding_smt}.")

    # If image as WCS, compute RA and Dec of each source
    try:
        # Retrieve the RA and Dec of each source as SKyCoord objects, then convert to
        # arrays of floats to add to table
        skypos = ccd.wcs.pixel_to_world(sources["xcentroid"], sources["ycentroid"])
        sources["ra"] = skypos.ra.value
        sources["dec"] = skypos.dec.value
    except AttributeError:
        # No WCS, so add empty columns
        sources["ra"] = np.nan * np.ones(src_cnt)
        sources["dec"] = np.nan * np.ones(src_cnt)

    # If requested, compute the FWHM of each source
    if find_fwhm:
        x, y = compute_fwhm(
            ccd,
            sources,
            fwhm_estimate=fwhm,
            x_column="xcentroid",
            y_column="ycentroid",
            sky_per_pix_avg=sky_per_pix_avg,
        )
        sources["fwhm_x"] = x
        sources["fwhm_y"] = y
        sources["width"] = (x + y) / 2  # Average of x and y FWHM

        # Flag bogus fwhm values returned from fitting (no objects
        # have a fwhm less than 1 pixel)
        bad_src = (sources["fwhm_x"] < 1) | (sources["fwhm_y"] < 1)
        sources["fwhm_x"][bad_src] = np.nan
        sources["fwhm_y"][bad_src] = np.nan
        sources["width"][bad_src] = np.nan
    else:  # add empty columns
        sources["fwhm_x"] = np.nan * np.ones(src_cnt)
        sources["fwhm_y"] = np.nan * np.ones(src_cnt)
        sources["width"] = np.nan * np.ones(src_cnt)

    # Convert sources to SourceListData object by adding
    # units to the columns
    units_dict = {
        "id": None,
        "ra": u.deg,
        "dec": u.deg,
        "xcentroid": u.pix,
        "ycentroid": u.pix,
        "fwhm_x": u.pix,
        "fwhm_y": u.pix,
        "width": u.pix,
    }
    sources = Table(data=sources, units=units_dict)
    # Rename columns to match SourceListData
    colnamemap = {"id": "star_id", "xcentroid": "xcenter", "ycentroid": "ycenter"}

    sl_data = SourceListData(input_data=sources, colname_map=colnamemap)
    return sl_data



import astropy.io.fits as fits
import numpy as np
from astropy.modeling.models import Gaussian2D
from astropy.nddata import CCDData
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS
from photutils.datasets import make_model_image, make_noise_image


class FakeImage:
    """
    Creates a fake image with a set of sources using the datafile stored
    at `data/test_sources.csv`.

    Parameters
    ----------

    noise_dev : float, optional
        The standard deviation of the noise in the image.

    seed : int, optional
        The seed to use for the random number generator. If not specified,
        the seed will be randomly generated.
    """

    def __init__(self, noise_dev=1.0, seed=None):
        self.image_shape = [400, 500]
        data_file = get_pkg_data_filename("data/test_sources.csv")
        self._sources = Table.read(data_file)
        self.mean_noise = self.sources["amplitude"].max() / 100
        self.noise_dev = noise_dev
        self._stars = make_gaussian_sources_image(tuple(self.image_shape), self.sources)
        self._noise = make_noise_image(
            self._stars.shape, mean=self.mean_noise, stddev=noise_dev, seed=seed
        )
        # Sky background per pixel should be the mean level of the noise.
        self._sources["sky_per_pix_avg"] = self.mean_noise

    @property
    def sources(self):
        """
        Return the table of sources used to create the fake image.
        """
        return self._sources

    @property
    def image(self):
        """
        Return the fake image.
        """
        return self._stars + self._noise


class FakeCCDImage(CCDData):
    # Generates a fake CCDData object for testing purposes.
    def __init__(self, *args, **kwargs):
        # Pull off the seed argument if it exists.
        seed = kwargs.pop("seed", None)

        # If no arguments are passed, use the default FakeImage.
        # This dodge is necessary because otherwise we can't copy the CCDData
        # object apparently.
        if (len(args) == 0) and (len(kwargs) == 0):
            base_data = FakeImage(seed=seed)
            super().__init__(base_data.image.copy(), unit="adu")

            # Append attributes from the base data object.
            self.sources = base_data.sources.copy()
            self.noise_dev = base_data.noise_dev
            self.mean_noise = base_data.mean_noise
            self.image_shape = base_data.image_shape.copy()

            # Add some additional features to the CCDData object, like
            # a header and the sources used to create the image.abs
            self.header = fits.Header()
            self.header["OBJECT"] = "Test Object"
            self.header["EXPOSURE"] = 1.0
            self.header["DATE-OBS"] = "2018-01-01T00:00:00.0"
            self.header["AIRMASS"] = 1.2
            self.header["FILTER"] = "V"

            # Set up a  WCS header for the CCDData object.
            (size_y, size_x) = base_data.image_shape
            pixel_scale = 0.75  # arcseconds per pixel
            ra_center = 283.6165
            dec_center = 33.05857
            w = WCS(naxis=2)
            w.wcs.crpix = [
                size_x / 2,
                size_y / 2,
            ]  # Reference pixel (center of the image)
            w.wcs.cdelt = [
                -pixel_scale / 3600,
                pixel_scale / 3600,
            ]  # Pixel scale in degrees per pixel
            w.wcs.crval = [
                ra_center,
                dec_center,
            ]  # RA and Dec of the reference pixel in degrees
            w.wcs.ctype = ["RA---TAN", "DEC--TAN"]  # Coordinate type (TAN projection)
            # Rotate image to be slightly off of horizontal
            north_angle_deg = 8.4
            w.wcs.pc = [
                [
                    np.cos(np.radians(north_angle_deg)),
                    -np.sin(np.radians(north_angle_deg)),
                ],
                [
                    np.sin(np.radians(north_angle_deg)),
                    np.cos(np.radians(north_angle_deg)),
                ],
            ]

            self.wcs = w
            self.header.update(w.to_header())

    def drop_wcs(self):
        # Convenience function to remove WCS information from the CCDData object
        # for testing purposes.
        self.wcs = None
        wcs_keywords = [
            "CTYPE",
            "CRPIX",
            "CRVAL",
            "CDELT",
            "CUNIT",
            "CD1_",
            "CD2_",
            "PC1_",
            "PC2_",
        ]
        for keyword in wcs_keywords:
            matching_keys = [key for key in self.header.keys() if keyword in key]
            for key in matching_keys:
                del self.header[key]


def shift_FakeCCDImage(ccd_data, x_shift, y_shift):
    """
    Create a test CCD image based on the first test image with the central
    positions shifted by the given amount.  To prevent any sources being
    shifted off the edge of the image, an error is thrown if this is
    happening.

    WCS information in the CCDData is also updated to reflect the shift.

    Parameters:
        ccd_data : `FakeCCDImage`
            The original CCDData object.

        x_shift : int
            The amount of shift in x direction (in pixels)

        y_shift : int
            The amount of shift in y direction (in pixels)

    Returns:
        `FakeCCDImage`: A new CCDData object.
    """
    # Copy WCS from original CCDData object
    shifted_ccd_data = ccd_data.copy()
    for key, val in ccd_data.__dict__.items():
        try:
            shifted_ccd_data.__dict__[key] = val.copy()
        except AttributeError:
            shifted_ccd_data.__dict__[key] = val
    shifted_wcs = ccd_data.wcs.deepcopy()

    # Calculate the new RA and Dec center after shifting
    x_shift = int(x_shift)
    y_shift = int(y_shift)
    ra_center_shifted, dec_center_shifted = shifted_wcs.all_pix2world(
        (shifted_ccd_data.data.shape[1]) / 2 + x_shift,
        (shifted_ccd_data.data.shape[0]) / 2 + y_shift,
        0,
    )

    # Shift source positions
    shifted_ccd_data.sources["x_mean"] -= x_shift
    shifted_ccd_data.sources["y_mean"] -= y_shift

    # Check shifted sources are still on the image
    if (
        (np.any(shifted_ccd_data.sources["x_mean"] < 0))
        | (np.any(shifted_ccd_data.sources["x_mean"] > shifted_ccd_data.data.shape[1]))
        | (np.any(shifted_ccd_data.sources["y_mean"] < 0))
        | (np.any(shifted_ccd_data.sources["y_mean"] > shifted_ccd_data.data.shape[0]))
    ):
        raise ValueError("Sources shifted off the edge of the image.")

    # Update the new RA and Dec center in the shifted WCS
    shifted_wcs.wcs.crval = [ra_center_shifted, dec_center_shifted]
    shifted_ccd_data.wcs = shifted_wcs

    # Make image
    srcs = make_gaussian_sources_image(
        tuple(shifted_ccd_data.image_shape), shifted_ccd_data.sources
    )
    background = make_noise_image(
        srcs.shape, mean=shifted_ccd_data.mean_noise, stddev=shifted_ccd_data.noise_dev
    )
    shifted_ccd_data.data = srcs + background

    return shifted_ccd_data


def make_gaussian_sources_image(shape, source_table, oversample=1.0):
    """
    Make an image containing 2D Gaussian sources.

    Parameters
    ----------
    shape : 2-tuple of int
        The shape of the output 2D image.

    source_table : `~astropy.table.Table`
        Table of parameters for the Gaussian sources. Each row of the
        table corresponds to a Gaussian source whose parameters are
        defined by the column names. With the exception of ``'flux'``,
        column names that do not match model parameters will be ignored
        (flux will be converted to amplitude). If both ``'flux'`` and
        ``'amplitude'`` are present, then ``'flux'`` will be ignored.
        Model parameters not defined in the table will be set to the
        default value.

    oversample : float, optional
        The sampling factor used to discretize the models on a pixel
        grid. If the value is 1.0 (the default), then the models will
        be discretized by taking the value at the center of the pixel
        bin. Note that this method will not preserve the total flux of
        very small sources. Otherwise, the models will be discretized by
        taking the average over an oversampled grid. The pixels will be
        oversampled by the ``oversample`` factor.

    Note
    ----

    The body of this function, including the docstring, are copy/pasted from
    photutils 1.13.0 because the function make_gaussian_sources_image was
    deprecated in that version. Link:

    https://github.com/astropy/photutils/blob/main/photutils/datasets/images.py#L388

    Returns
    -------
    image : 2D `~numpy.ndarray`
        Image containing 2D Gaussian sources.
    """
    model = Gaussian2D(x_stddev=1, y_stddev=1)

    if "x_stddev" in source_table.colnames:
        xstd = source_table["x_stddev"]
    else:
        xstd = model.x_stddev.value  # default
    if "y_stddev" in source_table.colnames:
        ystd = source_table["y_stddev"]
    else:
        ystd = model.y_stddev.value  # default

    colnames = source_table.colnames
    if "flux" in colnames and "amplitude" not in colnames:
        source_table = source_table.copy()
        source_table["amplitude"] = source_table["flux"] / (2.0 * np.pi * xstd * ystd)

    return make_model_image(
        shape,
        model,
        source_table,
        x_name="x_mean",
        y_name="y_mean",
        discretize_oversample=oversample,
    )



from astropy.utils.data import get_pkg_data_filename

from stellarphot.core import PhotometryData


# Why is this in a separate file?
#
# The bug we are trying to reproduce happens only when no objects from
# stellarphot.settings.models have been created. In test_core we create several of
# the objects, which adds to the table registry methods for reading our custom objects.
# Here we do a bare minimum of imports to avoid that.
def test_photometry_file_read():
    # Regression test for #408
    file_name = get_pkg_data_filename("data/test_photometry_data.ecsv")
    phot_data = PhotometryData.read(file_name)
    assert phot_data.camera is not None



import sys

import numpy as np
import pytest

from stellarphot.transit_fitting import TransitModelFit

batman = pytest.importorskip("batman")

DEFAULT_TESTING_PARAMS = dict(
    t0=2,
    period=5,
    duration=0.1,
    depth=10,
    inclination=90,
    airmass_trend=0.005,
    width_trend=0.01,
    spp_trend=0.002,
)

pytestmark = pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Intermittent hard crash on windows"
)


def _create_data_from_model_with_trends(transit_model, noise_dev=0.01):
    # Given a fully set up transit model, create some fake
    # data with a touch of noise.

    # Make our own batman model and data from it
    model = batman.TransitModel(transit_model._batman_params, transit_model.times)
    data = model.light_curve(transit_model._batman_params)

    if transit_model.airmass is not None:
        data += transit_model.model.airmass_trend * (transit_model.airmass)

    if transit_model.width is not None:
        data += transit_model.model.width_trend * transit_model.width

    if transit_model.spp is not None:
        data += transit_model.model.spp_trend * transit_model.spp

    if noise_dev > 0:
        # Make some noise
        generator = np.random.default_rng(432132)
        noise = generator.normal(scale=noise_dev, size=len(data))

        data += noise

    return data


def test_transit_fit_value_length_check():
    # Check that setting inconsistent lengths raises an error
    tmod = TransitModelFit()

    # Should work, all others are None
    tmod.times = np.array(list(range(5)))

    # Should work, new length is 5
    tmod.airmass = list(range(5))

    # Each of these should raise an error, has different length
    with pytest.raises(ValueError) as e:
        tmod.spp = list(range(3))
    assert "Length of spp not consistent" in str(e.value)

    with pytest.raises(ValueError) as e:
        tmod.width = list(range(3))
    assert "Length of width not consistent" in str(e.value)

    with pytest.raises(ValueError) as e:
        tmod.times = list(range(3))
    assert "Length of times not consistent" in str(e.value)

    with pytest.raises(ValueError) as e:
        tmod.airmass = list(range(3))
    assert "Length of airmass not consistent" in str(e.value)

    with pytest.raises(ValueError) as e:
        tmod.data = list(range(3))
    assert "Length of data not consistent" in str(e.value)


@pytest.mark.parametrize("detrend_by", ["airmass", "spp", "width"])
def test_setting_unsetting_parameters_updates_detrend(detrend_by):
    tmod = TransitModelFit()

    dummy_values = np.array(list(range(5)))
    # Should work, all others are None
    tmod.times = dummy_values

    setattr(tmod, detrend_by, dummy_values)
    assert detrend_by in tmod._detrend_parameters

    setattr(tmod, detrend_by, None)
    assert detrend_by not in tmod._detrend_parameters


def test_transit_fit_setting_independent_vars():
    tmod = TransitModelFit()

    values = np.array(list(range(5)))

    for attr in ["times", "airmass", "width", "spp"]:
        setattr(tmod, attr, values)
        assert (getattr(tmod, attr) == values).all()


def test_transit_fit_setting_independent_vars_to_none():
    """
    Setting to None should always work.
    """
    tmod = TransitModelFit()

    values = np.array(list(range(5)))

    # First set values to something that is not None
    for attr in ["times", "airmass", "width", "spp", "data"]:
        setattr(tmod, attr, values)

    # Now set to None and check value
    for attr in ["times", "airmass", "width", "spp", "data"]:
        setattr(tmod, attr, None)
        assert getattr(tmod, attr) is None


def test_transit_create_model():
    # Creating a model should not give us an error and should lead
    # to expected values of parameters.
    tmod = TransitModelFit()

    tmod.setup_model(**DEFAULT_TESTING_PARAMS)

    duration = DEFAULT_TESTING_PARAMS["duration"]
    period = DEFAULT_TESTING_PARAMS["period"]

    expected_a = 1 / np.sin(duration * np.pi / period)
    assert tmod.model.t0 == DEFAULT_TESTING_PARAMS["t0"]
    assert tmod.model.rp == np.sqrt(DEFAULT_TESTING_PARAMS["depth"] / 1000)
    assert np.abs(tmod.model.a - expected_a) < 1e-7
    assert tmod.model.period == DEFAULT_TESTING_PARAMS["period"]
    assert tmod.model.airmass_trend == DEFAULT_TESTING_PARAMS["airmass_trend"]
    assert tmod.model.width_trend == DEFAULT_TESTING_PARAMS["width_trend"]
    assert tmod.model.spp_trend == DEFAULT_TESTING_PARAMS["spp_trend"]


def _make_transit_model_with_data(
    noise_dev=1e-1, with_airmass=False, with_width=False, with_spp=False
):
    tmod = TransitModelFit()

    tmod.setup_model(**DEFAULT_TESTING_PARAMS)

    t0 = DEFAULT_TESTING_PARAMS["t0"]
    duration = DEFAULT_TESTING_PARAMS["duration"]

    # Last factor ensures some out of transit data
    start = t0 - duration / 2 * 1.3
    end = t0 + duration / 2 * 1.3

    times = np.linspace(start, end, num=100)
    tmod.times = times

    # Make an airmass that starts at 1.4 and decreases to 1.0
    # over the duration and is parabolic centered on the final time.
    airmass = 0.4 / (times[0] - times[-1]) ** 2 * (times - times[-1]) ** 2 + 1.0

    tmod.airmass = airmass if with_airmass else None

    # Make a width that increases linearly with time. Never happen IRL, but
    # that is ok. Let's start at 5 pixels and increase to 9.
    width = 4 * (times - times[0]) / (times[-1] - times[0]) + 5

    tmod.width = width if with_width else None

    # sky_per_pixel...let's make that sinusoidal. Not realistic but should be
    # easy for the fitter to pick out this trend.
    spp = 30 + 5 * np.sin(4 * np.pi * times / (times[-1] - times[0]))

    tmod.spp = spp if with_spp else None

    data = _create_data_from_model_with_trends(tmod, noise_dev=noise_dev)

    tmod.data = data

    return tmod


def test_transit_fit_all_parameters():
    tmod = _make_transit_model_with_data(
        noise_dev=1e-5, with_airmass=True, with_width=True, with_spp=True
    )

    duration = DEFAULT_TESTING_PARAMS["duration"]
    period = DEFAULT_TESTING_PARAMS["period"]

    expected_a = 1 / np.sin(duration * np.pi / period)

    assert np.abs(tmod.model.a - expected_a) < 1e-6

    tmod.fit()

    # Check the non-exoplanet trends
    for fit_trend in ["airmass_trend", "width_trend", "spp_trend"]:
        assert (
            np.abs(getattr(tmod.model, fit_trend) - DEFAULT_TESTING_PARAMS[fit_trend])
            < 1e-5
        )

    # Check a few of the exoplanet parameters
    assert np.abs(tmod.model.t0 - DEFAULT_TESTING_PARAMS["t0"]) < 1e-3

    assert np.abs(tmod.model.a - expected_a) < 1e-2

    expected_rp = np.sqrt(DEFAULT_TESTING_PARAMS["depth"] / 1000)
    assert np.abs(tmod.model.rp - expected_rp) < 1e-4

    assert "airmass" in tmod._detrend_parameters
    assert "width" in tmod._detrend_parameters
    assert "spp" in tmod._detrend_parameters


def test_transit_model_detrend():
    tmod = _make_transit_model_with_data(
        noise_dev=0, with_airmass=True, with_width=True, with_spp=True
    )

    no_trends = _make_transit_model_with_data(
        noise_dev=0, with_airmass=False, with_width=False, with_spp=False
    )
    full_model = tmod.model_light_curve()

    np.testing.assert_allclose(
        full_model - no_trends.model_light_curve(),
        tmod.model.airmass_trend * tmod.airmass
        + tmod.model.width_trend * tmod.width
        + tmod.model.spp_trend * tmod.spp,
    )

    for trend in tmod._all_detrend_params:
        detrended_model = tmod.model_light_curve(detrend_by=trend)
        trend_param = getattr(tmod.model, f"{trend}_trend")
        trend_data = getattr(tmod, trend)
        assert trend_param != 0
        assert trend_data is not None
        np.testing.assert_allclose(
            full_model - detrended_model, trend_param * trend_data
        )

    detrended_model = tmod.model_light_curve(detrend_by="all")
    np.testing.assert_allclose(detrended_model, no_trends.model_light_curve())


def test_transit_data_detrend():
    tmod = _make_transit_model_with_data(
        noise_dev=0, with_airmass=True, with_width=True, with_spp=True
    )

    no_trends = _make_transit_model_with_data(
        noise_dev=0, with_airmass=False, with_width=False, with_spp=False
    )

    np.testing.assert_allclose(tmod.data_light_curve(detrend_by="all"), no_trends.data)
    np.testing.assert_allclose(
        tmod.data - no_trends.data,
        tmod.model.airmass_trend * tmod.airmass
        + tmod.model.width_trend * tmod.width
        + tmod.model.spp_trend * tmod.spp,
    )


def test_transit_fit_parameters_unfreeze_as_expected():
    tmod = _make_transit_model_with_data(
        noise_dev=1e-5, with_airmass=False, with_width=False, with_spp=False
    )

    # None of these are fixed by default
    assert not tmod.model.airmass_trend.fixed
    assert not tmod.model.width_trend.fixed
    assert not tmod.model.spp_trend.fixed

    tmod.fit()

    # They should return to their original state after the
    # fit even though they are temporarily fixed during the
    # fit.
    assert not tmod.model.airmass_trend.fixed
    assert not tmod.model.width_trend.fixed
    assert not tmod.model.spp_trend.fixed



import json
import re
from pathlib import Path

import ipywidgets as ipw
import numpy as np
from astropy.utils.data import get_pkg_data_filename
from traitlets import Bool, observe

from stellarphot.transit_fitting.io import get_tic_info

__all__ = [
    "MyValid",
    "make_checker",
    "validate_exposure_time",
    "populate_TIC_boxes",
    "populate_TOI_boxes",
    "exotic_settings_widget",
    "set_values_from_json_file",
    "get_values_from_widget",
    "generate_json_file_name",
]


template_types = ["known", "candidate"]
template_json = {}
to_fill = {}

for template in template_types:
    template_name = get_pkg_data_filename(
        "data/tic-template-for-exotic-" f"{template}.json"
    )
    with open(template_name) as f:
        template_json[template] = json.load(f)

    template_name = get_pkg_data_filename(f"data/exotic-to-mod-{template}.json")
    with open(Path(template_name)) as f:
        to_fill[template] = json.load(f)

exotic_arguments = dict(
    known=["--nasaexoarch", "--pre"], candidate=["--override", "--pre"]
)

# Nested keys are flattened by joining them with this character
join_char = "😬"


class MyValid(ipw.Button):
    """
    A class containing a more compact indicator of valid entries based on ipywidgets
    button value.

    Parameters
    ----------

    **kwd :
        All keyword arguments used to initialize an instance of this class are
        passed to the ipywidgets.Button constructor.

    Attributes
    ----------

    value : bool
        Current value of the indicator. Initialized to False.

    """

    value = Bool(False, help="Bool value").tag(sync=True)

    def __init__(self, **kwd):
        super().__init__(**kwd)
        self.layout.width = "40px"
        self._set_properties(None)

    @observe("value")
    def _set_properties(self, change):  # noqa: ARG002
        """
        Widget callbacks need to accept a single argument, even if it is not used.
        """
        if self.value:
            self.style.button_color = "green"
            self.icon = "check"
        else:
            self.style.button_color = "red"
            self.icon = "times"


def make_checker(indicator_widget, value_widget):
    """
    Build an observer that checks TIC number and, if it is a valid
    TIC number, looks up information about the TIC object from
    MAST copy of TIC as priors for EXOTIC. It also sets the visible checkbox
    to the appropriate state.

    Parameters
    ----------

    indicator_widget : `~stellarphot.transit_fitting.gui.MyValid` widget
        The widget that indicates to the user whether or not the value is
        reasonable.

    value_widget: ipywidget
        This widget should be generated by exotic_settings_widget.

    Returns
    -------

    function
        Function with the correct signature for use as an observer on an
        ipywidget.
    """

    def check_name(change):
        # Valid TIC number is 9 digits
        ticced = re.compile(r"TIC \d{9,10}$")
        owner = change["owner"]
        is_tic = ticced.match(change["new"])
        if is_tic:
            if indicator_widget is not None:
                indicator_widget.value = True
            owner.disabled = True
            tic_info = get_tic_info(change["new"][-9:])
            if not tic_info:
                indicator_widget.value = False
                indicator_widget.tooltip = "Not a valid TIC number"
            else:
                populate_TIC_boxes(tic_info, value_widget)
            owner.disabled = False
        else:
            owner.disabled = False
            if indicator_widget is not None:
                indicator_widget.value = False
                indicator_widget.tooltip = "TIC numbers have 9 digits"

    return check_name


def validate_exposure_time(
    indicator_widget, value_widget  # noqa: ARG001 (value_widget needed for callback)
):
    """Validates the exposure time input.

    Parameters
    ----------
    indicator_widget : `~stellarphot.transit_fitting.gui.MyValid` widget
        The widget that indicates to the user whether or not the value is
        reasonable.

    value_widget: ipywidget
        This widget should be generated by exotic_settings_widget.

    Returns
    -------
    function
        Function that will set the correct boolean value on the
        indicator_widget to indicate if the value of the exposure time
        is valid.  This can be used as an observer for an ipywidget
    """

    def check_exposure(change):
        # Valid Exposure time is greater than zero
        if change["new"] > 0:
            if indicator_widget is not None:
                indicator_widget.value = True
        else:
            if indicator_widget is not None:
                indicator_widget.value = False

    return check_exposure


def populate_TIC_boxes(tic_info, value_widget):
    """
    Set the appropriate widget values given information pulled from
    MAST TIC.

    Parameters
    ----------

    tic_info : dict
        Dictionary of information about the TIC object.

    value_widget: ipywidget
        This widget should contain a 'candidate' key.

    Returns
    -------

    None
        Sets values of planetary parameters of `candidate` in ``value_widget`` in place.
    """
    # Match EXOTIC json keys to columns in the result returned from
    # astroquery
    exotic_tic = {
        "Star Effective Temperature (K)": "Teff",
        "Star Effective Temperature (+) Uncertainty": "epos_Teff",
        "Star Effective Temperature (-) Uncertainty": "eneg_Teff",
        "Star Surface Gravity (log(g))": "logg",
        "Star Surface Gravity (+) Uncertainty": "epos_logg",
        "Star Surface Gravity (-) Uncertainty": "eneg_logg",
        "Host Star Name": "UCAC",
        "Star Metallicity ([FE/H])": "MH",
        "Star Metallicity (+) Uncertainty": "e_MH",
        "Star Metallicity (-) Uncertainty": "e_MH",
    }
    for k, v in exotic_tic.items():
        exotic_key = join_char.join(["planetary_parameters", k])
        if k == "Host Star Name":
            value_widget["candidate"][exotic_key].value = f"UCAC4 {tic_info[v][0]}"
        elif not np.isnan(tic_info[v][0]):
            value_widget["candidate"][exotic_key].value = tic_info[v][0]


def populate_TOI_boxes(toi, exotic_widget):
    """
    Set the appropriate widget values given information pulled from
    TESS Target of Interest (TOI) list.

    Parameters
    ----------

    toi : `~stellarphot.io.tess.TOI`
        Information about the TESS Target of Interest (TOI) object.

    exotic_widget:  ipywidget
        This widget should be generated by exotic_settings_widget.

    Returns
    -------

    None
        Sets values of planetary parameters of `candidate` in
        ``exotic_widget`` in place.
    """
    # Match EXOTIC json keys to columns in the result returned from
    # astroquery
    exotic_toi = {
        "Planet Name": "tic_id",
        "Target Star RA": "coord",
        "Target Star Dec": "coordP",
        "Orbital Period (days)": "period",
        "Orbital Period Uncertainty": "period_error",
        "Published Mid-Transit Time (BJD-UTC)": "epoch",
        "Mid-Transit Time Uncertainty": "epoch_error",
        # Could maybe get these from TOI information, but not straightforward
        # "Ratio of Planet to Stellar Radius (Rp/Rs)": 0.0,
        # "Ratio of Planet to Stellar Radius (Rp/Rs) Uncertainty": 0.0,
        # "Ratio of Distance to Stellar Radius (a/Rs)": 0.0,
        # "Ratio of Distance to Stellar Radius (a/Rs) Uncertainty": 0.0,
    }
    for k, v in exotic_toi.items():
        exotic_key = join_char.join(["planetary_parameters", k])
        if k == "Planet Name":
            exotic_widget["candidate"][exotic_key].value = f"TIC {toi.tic_id}"
        elif k == "Target Star RA":
            exotic_widget["candidate"][exotic_key].value = toi.coord.ra.to_string(
                unit="hour", decimal=False, sep=":"
            )
        elif k == "Target Star Dec":
            exotic_widget["candidate"][exotic_key].value = toi.coord.dec.to_string(
                unit="degree", decimal=False, sep=":"
            )
        else:
            exotic_widget["candidate"][exotic_key].value = getattr(toi, v).value


# This sets up the specific widgets whose values get validated.
# That includes the widget that contains the TIC number for candidates.
validators = dict(known={}, candidate={})
validators["candidate"]["Planet Name"] = make_checker
for k in validators:
    validators[k]["Exposure Time (s)"] = validate_exposure_time


def exotic_settings_widget():
    """
    Generate a widget to enter (and store) settings for exotic.

    Parameters
    ----------

    None

    Returns
    -------

    ipywidget
        Widget with settings for EXOTIC.
    """

    # We rely on some global variables:
    global to_fill, template_types

    # This dictionary will contain all of the widgets
    widget_list = {}

    # Each widget has the same layout for its description and its value/input
    layout_description = ipw.Layout(width="70%")
    layout_input = ipw.Layout(width="30%")

    # Maintain a separate dict of just the value widgets
    value_widget = {}

    # For exotic there are two templates, one for known exoplanets and one for
    # candidate exoplanets
    for template in template_types:
        # Keep separate track of sidgets that make up each template
        value_widget[template] = {}
        widget_list[template] = []
        # Autogenerate the input fields and labels
        for k in to_fill[template]:
            for k2, v in to_fill[template][k].items():
                if isinstance(v, str):
                    input_widget = ipw.Text(value=v)
                elif isinstance(v, float):
                    if v >= 0:
                        # In principle we could limit this to positive values
                        input_widget = ipw.FloatText(value=v)
                    else:
                        input_widget = ipw.FloatText(value=v)
                input_widget.layout = layout_input

                # Each horizontal box below is one "cell" and the input grid.
                # The HTML widget has the label, and input_widget is the text or
                # float widget.
                hb = ipw.HBox(
                    [ipw.HTML(value=k2, layout=layout_description), input_widget]
                )

                # Widgets with validators need some observers added
                try:
                    # validator is a function that generates a function with the right
                    # with the right signature to handle an ipywidgets event
                    validator = validators[template][k2]
                except KeyError:
                    # Most of the widgets...
                    pass
                else:
                    kids = list(hb.children)
                    # Add a visible indication of whether the value is reasonable
                    indicator = MyValid(value=False)
                    kids.append(indicator)

                    # Add an observer to the widget to watch for changes
                    input_widget.observe(
                        validator(indicator, value_widget), names="value"
                    )

                    hb.children = kids

                val_key = join_char.join([k, k2])
                value_widget[template][val_key] = input_widget
                widget_list[template].append(hb)

    hb2 = {}
    for template in template_types:
        hb2[template] = ipw.HBox(
            [
                ipw.VBox(widget_list[template][:16], layout=ipw.Layout(padding="10px")),
                ipw.VBox(widget_list[template][16:]),
            ]
        )

    select_planet_type = ipw.ToggleButtons(
        description="Known or candidate exoplanet?",
        options=template_types,
        style={"description_width": "initial"},
    )

    lookup_link_text = dict(
        known="https://exoplanetarchive.ipac.caltech.edu/",
        candidate="https://exofop.ipac.caltech.edu/tess/",
    )

    lookup_link_html = {}

    for k, v in lookup_link_text.items():
        lookup_link_html[k] = ipw.HTML(
            f"<h3>For some information about this "
            f'object: <a href="{v}" target="_blank">{v}</a></h3>'
        )

    input_container = ipw.VBox()

    whole_thing = ipw.VBox(children=[select_planet_type, input_container])
    whole_thing.planet_type = select_planet_type
    whole_thing.value_widget = value_widget
    pre_reduced_file = join_char.join(["optional_info", "Pre-reduced File:"])
    whole_thing.data_file_widget = {
        "candidate": value_widget["candidate"][pre_reduced_file],
        "known": value_widget["known"][pre_reduced_file],
    }

    def observe_select(change):  # noqa: ARG001
        """
        Widget callbacks need to accept a single argument, even if it is not used.
        """
        input_container.children = [
            lookup_link_html[select_planet_type.value],
            hb2[select_planet_type.value],
        ]

    select_planet_type.observe(observe_select, names="value")
    observe_select(select_planet_type.value)

    return whole_thing


def set_values_from_json_file(widget, json_file):
    """
    Set values in EXOTIC widget from a JSON file of those settings.

    Parameters
    ----------

    widget : ipywidget
        This widget should be generated by exotic_settings_widget.

    json_file : str
        File with settings for the widget.

    Returns
    -------
    None
        Sets values of parameters of widget in place.
    """
    with open(json_file) as f:
        input_values = json.load(f)

    planet_type = widget.planet_type.value
    for k, a_widget in widget.value_widget[planet_type].items():
        k1, k2 = k.split(join_char)
        a_widget.value = input_values[k1][k2]


def get_values_from_widget(exotic_widget, key=None):
    """
    Extract EXOTIC settings from widget.

    Parameters
    ----------

    exotic_widget : ipywidget
        This widget should be generated by exotic_settings_widget.

    key : str, either "known" or "candidate", optional
        Indicates which case to use for EXOTIC. If ``None``, use the ``planet_type``
        attribute of the ``exotic_widget``.

    Returns
    -------

    dict
        Value of modified template_json[key], where template_json is a global variable.

    """
    if not key:
        key = exotic_widget.planet_type.value

    for k, widget in exotic_widget.value_widget[key].items():
        k1, k2 = k.split(join_char)
        template_json[key][k1][k2] = widget.value

    return template_json[key]


def generate_json_file_name(exotic_widget, key=None):
    """
    Generate settings file name from user input.

    Parameters
    ----------

    exotic_widget : ipywidget
        This widget should be generated by exotic_settings_widget.

    key : str, either "known" or "candidate", optional
        Indicates which case to use for EXOTIC. If ``None``, use the ``planet_type``
        attribute of the ``exotic_widget``.

    Returns
    -------
    str
        Name of file to save settings to.
    """
    if not key:
        key = exotic_widget.planet_type.value

    get_values_from_widget(exotic_widget, key=key)
    user_info = "user_info"
    planet = "planetary_parameters"
    filter_key = "Filter Name (aavso.org/filters)"
    date = template_json[key][user_info]["Observation date"]
    planet_name = template_json[key][planet]["Planet Name"]
    filter_name = template_json[key][user_info][filter_key]
    name = f"{planet_name}-{date}-{filter_name}"
    return name.replace(" ", "_") + ".json"



import re
from dataclasses import dataclass
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Annotated

import requests
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.time import Time
from astropy.utils.data import download_file
from pydantic import BaseModel, ConfigDict

from stellarphot import SourceListData
from stellarphot.settings.astropy_pydantic import (
    AstropyValidator,
    QuantityType,
    WithPhysicalType,
)
from stellarphot.transit_fitting.io import get_tic_info

__all__ = ["tess_photometry_setup", "TessSubmission", "TOI", "TessTargetFile"]

# Makes me want to vomit, but....
DEFAULT_TABLE_LOCATION = "who.the.heck.knows"
TOI_TABLE_URL = "https://exofop.ipac.caltech.edu/tess/download_toi.php?output=csv"
GAIA_APERTURE_SERVER = "https://www.astro.louisville.edu/"
TIC_regex = re.compile(r"[tT][iI][cC][^\d]?(?P<star>\d+)(?P<planet>\.\d\d)?")

MODEL_DEFAULT_CONFIGURATION = ConfigDict(
    # Make sure default values are valid
    validate_default=True,
    # Make sure changes to values made after initialization are valid
    validate_assignment=True,
    # Make sure there are no extra fields
    extra="forbid",
)


@dataclass
class TessSubmission:
    """
    A data class to represent TESS submissions.

    Parameters
    ----------

    telescope_code : str
        The telescope code, e.g. "SRO" or "TJO"

    filter : str
        The filter used for the observations, e.g. "Ic" or "Rc"

    utc_start : str
        The UTC date of the first observation, in YYYYMMDD format

    tic_id : int
        The TIC ID of the target

    planet_number : int
        The planet number, if applicable

    Attributes
    ----------

    apertures : str

    base_name : str

    field_image : str

    field_image_zoom : str

    filter: str
        The filter used for the observations, e.g. "Ic" or "Rc"

    planet_number : int
        The planet number, if applicable

    seeing_profile : str

    tic_coord : `astropy.coordinates.SkyCoord`

    tic_id : int
        The TIC ID of the target

    telescope_code : str
        The telescope code, e.g. "SRO" or "TJO"

    utc_start : str
        The UTC date of the first observation, in YYYYMMDD format

    """

    telescope_code: str
    filter: str
    utc_start: int
    tic_id: int
    planet_number: int

    def __post_init__(self, *args, **kwargs):
        self._tic_info = None

    @classmethod
    def from_header(cls, header, telescope_code="", planet=0):
        """
        Create a TessSubmission from a FITS header

        Parameters
        ----------
        header : `astropy.io.fits.Header`
            The FITS header to parse

        telescope_code : str
            The telescope code, e.g. "SRO" or "TJO"

        planet : int
            The planet number, if applicable
        """

        tic_id = 0
        filter = ""
        fails = {}
        try:
            dateobs = header["date-obs"]
        except KeyError:
            fails["utc_start"] = "UTC date of first image"
        else:
            dateobs = dateobs.split("T")[0].replace("-", "")

        try:
            filter = header["filter"]
        except KeyError:
            fails["filter"] = "filter/passband"

        try:
            obj = header["object"]
        except KeyError:
            fails["tic_id"] = "TIC ID number"
        else:
            result = TIC_regex.match(obj)
            if result:
                tic_id = int(result.group("star"))
                # Explicit argument overrules the header
                if result.group("planet") and not planet:
                    # Drop the leading period from the match
                    planet = int(result.group("planet")[1:])
            else:
                # No star from the object after all
                fails["tic_id"] = "TIC ID number"

        fail = []
        for k, v in fails.items():
            fail.append(f"Unable to determine {k}, {v}, from header.")

        fail = "\n".join(fail)

        if fail:
            raise ValueError(fail)

        return cls(
            utc_start=dateobs,
            filter=filter,
            telescope_code=telescope_code,
            tic_id=tic_id,
            planet_number=planet,
        )

    def _valid_tele_code(self):
        return len(self.telescope_code) > 0

    def _valid_planet(self):
        return self.planet_number > 0

    def _valid_tic_num(self):
        return self.tic_id < 10_000_000_000

    def _valid(self):
        """
        Check whether the information so far is valid, meaning:
         + Telescope code is not the empty string
         + Planet number is not zero
         + TIC ID is not more than 10 digits
        """
        valid = (
            self._valid_tele_code() and self._valid_planet() and self._valid_tic_num()
        )
        return valid

    @property
    def base_name(self):
        """
        The base name of the submission, e.g. "TIC123456789-01_20200101_SRO_Ic"
        """
        if self._valid():
            pieces = [
                f"TIC{self.tic_id}-{self.planet_number:02d}",
                self.utc_start,
                self.telescope_code,
                self.filter,
            ]
            return "_".join(pieces)

    @property
    def seeing_profile(self):
        """
        The name of the seeing profile file,
        e.g. "TIC123456789-01_20200101_SRO_Ic_seeing-profile.png"
        """
        return self.base_name + "_seeing-profile.png"

    @property
    def field_image(self):
        """
        The name of the field image file,
        e.g. "TIC123456789-01_20200101_SRO_Ic_field.png"
        """
        return self.base_name + "_field.png"

    @property
    def field_image_zoom(self):
        """
        The name of the zoomed-in field image file,
        e.g. "TIC123456789-01_20200101_SRO_Ic_field-zoom.png"
        """
        return self.base_name + "_field-zoom.png"

    @property
    def apertures(self):
        """
        The name of the apertures file,
        e.g. "TIC123456789-01_20200101_SRO_Ic_measurements.apertures"
        """
        return self.base_name + "_measurements.apertures"

    @property
    def tic_coord(self):
        """
        The SkyCoord of the target, from the TIC catalog.
        """
        if not self._tic_info:
            self._tic_info = get_tic_info(self.tic_id)
        return SkyCoord(
            ra=self._tic_info["ra"][0], dec=self._tic_info["dec"][0], unit="degree"
        )

    def invalid_parts(self):
        """
        Prints a string identifying parts of the submission that are invalid.
        If submission valid, returns nothing.
        """
        if self._valid():
            return

        if not self._valid_tele_code():
            print(f"Telescope code {self.telescope_code} is not valid")

        if not self._valid_planet():
            print(f"Planet number {self.planet_number} is not valid")

        if not self._valid_tic_num():
            print(f"TIC ID {self.tic_id} is not valid.")


class TOI(BaseModel):
    """
    Represent a TOI from the ExoFOP database.

    Parameters
    ----------

    tic_id : int
        The TIC ID of the target.

    coord : `astropy.coordinates.SkyCoord`
        The coordinates of the target.

    depth_ppt : float
        The depth of the transit in parts per thousand.

    depth_error_ppt : float
        The error in the depth of the transit in parts per thousand.

    duration : `astropy.units.Quantity`
        The duration of the transit; must have units of time.

    duration_error : `astropy.units.Quantity`
        The error in the duration of the transit; must have units of time.

    epoch : `astropy.time.Time`
        The epoch of the transit.

    epoch_error : `astropy.units.Quantity`
        The error in the epoch of the transit; must have units of time.

    period : `astropy.units.Quantity`
        The period of the transit; must have units of time.

    period_error : `astropy.units.Quantity`
        The error in the period of the transit; must have units of time.

    tess_mag : float
        The TESS magnitude of the target.

    tess_mag_error : float
        The error in the TESS magnitude of the target.
    """

    model_config = MODEL_DEFAULT_CONFIGURATION

    tic_id: int
    coord: Annotated[SkyCoord, AstropyValidator]
    depth_ppt: float
    depth_error_ppt: float
    duration: Annotated[QuantityType, WithPhysicalType("time")]
    duration_error: Annotated[QuantityType, WithPhysicalType("time")]
    epoch: Annotated[Time, AstropyValidator]
    epoch_error: Annotated[QuantityType, WithPhysicalType("time")]
    period: Annotated[QuantityType, WithPhysicalType("time")]
    period_error: Annotated[QuantityType, WithPhysicalType("time")]
    tess_mag: float
    tess_mag_error: float

    @classmethod
    def from_tic_id(cls, tic_id):
        """
        Create a TOI object from a numerical TIC ID number. This will be obtained
        from ExoFOP-TESS and the TESS Input Catalog (TIC) at MAST.
        """
        toi_table = download_file(
            TOI_TABLE_URL, cache=True, show_progress=True, timeout=60
        )
        toi_table = Table.read(toi_table, format="ascii.csv")
        toi_table = toi_table[toi_table["TIC ID"] == tic_id]
        if len(toi_table) != 1:  # pragma: no cover
            raise RuntimeError(f"Found {len(toi_table)} rows in table, expected one.")
        toi_table = toi_table[0]

        # Retrieve some additional information from the TIC catalog at MAST, and grab
        # the first row of the table.
        tic_info = get_tic_info(tic_id)[0]

        return cls(
            tic_id=tic_id,
            coord=SkyCoord(ra=tic_info["ra"], dec=tic_info["dec"], unit="degree"),
            depth_ppt=toi_table["Depth (ppm)"] / 1000,
            depth_error_ppt=toi_table["Depth (ppm) err"] / 1000,
            duration=toi_table["Duration (hours)"] * u.hour,
            duration_error=toi_table["Duration (hours) err"] * u.hour,
            epoch=Time(toi_table["Epoch (BJD)"], scale="tdb", format="jd"),
            epoch_error=toi_table["Epoch (BJD) err"] * u.day,
            period=toi_table["Period (days)"] * u.day,
            period_error=toi_table["Period (days) err"] * u.day,
            tess_mag=toi_table["TESS Mag"],
            tess_mag_error=toi_table["TESS Mag err"],
        )


def tess_photometry_setup(tic_id=None, TOI_object=None, overwrite=False):
    """
    Set up the photometry for a TESS target.

    Parameters
    ----------

    TIC_ID : int, optional
        The TIC ID of the target. Must provide either this or TOI_object.

    TOI_object : `stellarphot.io.TOI`, optional
        The TOI object for the target. Must provide either this or TIC_ID.

    Returns
    -------

    Nothing. Writes files with the TESS target information.
    """
    if tic_id:
        toi = TOI.from_tic_id(tic_id)
    elif TOI_object:
        toi = TOI_object
    else:
        raise ValueError("Must provide either TIC ID or TOI object.")

    ttf = TessTargetFile(toi.coord, toi.tess_mag, toi.depth_ppt)
    new_table = Table(
        {
            "star_id": list(range(len(ttf.table))),
            "ra": ttf.table["coords"].ra,
            "dec": ttf.table["coords"].dec,
            "marker name": ["TESS Targets"] * len(ttf.table),
            "coords": ttf.table["coords"],
        }
    )
    sld = SourceListData(input_data=new_table)
    sl_name = f"TIC-{toi.tic_id}-source-list-input.ecsv"
    try:
        sld.write(sl_name, overwrite=overwrite)
    except OSError as e:
        raise FileExistsError(
            f"Source list {sl_name} already exists: Use overwrite=True to replace"
        ) from e

    info_path = Path(f"TIC-{toi.tic_id}-info.json")
    if info_path.exists() and not overwrite:
        raise FileExistsError(
            f"{info_path} already exists. Use overwrite=True to replace."
        )
    with open(info_path, "w") as f:
        f.write(toi.model_dump_json(indent=2))


@dataclass
class TessTargetFile:
    """
    A class to hold information about a TESS target file.  It will retrieve all
    GAIA EDR3 sources within 2.5 arcminutes of the target using the online service at:
    https://www.astro.louisville.edu/gaia_to_aij/index.html

    Parameters
    ----------

    coord : `astropy.coordinates.SkyCoord`
        The coordinates of the target.

    magnitude : float
        The magnitude of the target.

    depth : float
        The depth of the transit.

    file : str, optional
        The path to a file that will be written containing the GAIA sources
        within 2.5 arcminutes the TESS target. If not provided, a temporary
        file will be created.

    Attributes
    ----------

    aperture_server : str
        The URL of the aperture server.

    coord : `astropy.coordinates.SkyCoord`
        The coordinates of the target.

    depth : float
        The depth of the transit.

    file : str
        The path to the file to create with the downloaded GAIA
        data. If not provided, a temporary file will be created.

    magnitude : float
        The magnitude of the target.

    table : `astropy.table.Table`
        A table of targets read in from target_file.

    target_file : str
        The path to the target file.

    target_table : `astropy.table.Table`
        The target table.

    """

    coord: SkyCoord
    magnitude: float
    depth: float
    file: str = ""

    def __post_init__(self):
        self.aperture_server = GAIA_APERTURE_SERVER
        if not self.file:
            # Be sure not to delete the file -- otherwise, on windows,
            # the file is deleted immediately as far as I can tell.
            self.file = NamedTemporaryFile(delete=False)
        self._path = Path(self.file.name)
        self.target_file = self._retrieve_target_file()
        self.target_table = self._build_table()

    def _retrieve_target_file(self):
        params = dict(
            ra=self.coord.ra.to_string(unit="hour", decimal=False, sep=":"),
            dec=self.coord.dec.to_string(unit="degree", decimal=False, sep=":"),
            mag=self.magnitude,
            depth=self.depth,
        )
        result = requests.get(
            self.aperture_server + "cgi-bin/gaia_to_aij/upload_request.cgi",
            params=params,
            timeout=15,  # If no response in 15 seconds we won't ever get one...
        )
        if result.status_code != 200:
            raise requests.ConnectionError(
                f"Failed to retrieve target file: {result.text}"
            )

        links = re.search(
            'href="(.+)"',
            result.text.replace("\n", ""),
        )
        download_link = self.aperture_server + links[1]
        target_file_contents = requests.get(download_link)
        # Write GAIA data to local file
        with open(self._path, "w") as f:
            f.write(target_file_contents.text)

    def _build_table(self):
        from stellarphot.utils.comparison_utils import read_file

        self.table = read_file(self._path)



from ipyautoui import AutoUi
from ipywidgets import Layout

from .models import SourceLocationSettings, _extract_short_description

__all__ = ["ui_generator"]


def ui_generator(model, max_field_width=None, file_chooser_max_width=None):
    """
    Generate a user interface with ipyautoui with a few default settings.

    Parameters
    ----------
    model : `pydantic.BaseModel` subclass
        The model to generate the user interface for.

    max_field_width : str, optional
        The width of the fields in the user interface. Default is `None`, which
        will use the default width of the fields. The value is passed on to the
        `layout.width` attribute of the fields, which can be any valid CSS width.
        These typically include units, e.g. "100px" or "10em".

    file_chooser_max_width : str, optional
        The width of the file chooser fields in the user interface. Default is `None`,
        which will use the default width of the fields. This is separate from
        `max_field_width` because the FileChooser widget uses a different layout that
        is less cluttered than the fields.

    """
    ui = AutoUi(model)

    # Oof, there is a bug in the file choose from ipyautoui. We patch up the
    # SourceLocationSettings widget here.
    if model == SourceLocationSettings:

        # Force a call to the handler to make sure the initial value is correct
        name = model.model_fields["source_list_file"].default
        value = ui.value.copy()
        value["source_list_file"] = name
        ui.value = value

    # validation is really messy looking right now, so suppress display of
    # the validation errors. A green check or red x will still be displayed.
    ui.show_validation = False

    # By default nullable are not shown at all but it seems much easier for
    # the user to understand if they are shown but disabled.
    ui.show_null = True

    # In the same spirit, the button to show/hide nullables should be hidden
    # too.
    ui.bn_shownull.layout.display = "none"

    # Set the description to the first sentence of the docstring. The default is
    # to use the entire docstring, which is often too long.
    ui.description = _extract_short_description(model.__doc__)

    # Always show nested models
    ui.open_nested = True

    # Validation is checked every time a value is changed, and the contents of each
    # field are written to the widget if validation passes. In at least one case,
    # the "Observatory" model, this is not helpful because the format of lat/lon is
    # not necessarily what the user entered. Furthermore, you can't edit the value
    # in the widget because it is overwritten every time validation passes. So, for
    # now, we will disable this feature by turning continuous update off for most
    # fields.
    for widget in ui.di_widgets.values():
        if hasattr(widget, "continuous_update"):
            widget.continuous_update = False

    # In some cases, the entry fields are too wide to fit in the space available, which
    # makes the widget look bad. We can set the width of the fields to a fixed value
    # to make them fit better.
    for widget in ui.di_widgets.values():
        if widget.__class__.__name__ == "FileChooser":
            if file_chooser_max_width is not None:
                # In a surprising twist, all FileChooser widgets seem to use the same
                # Layout object under the hood. So, if we change the width of one, we
                # change the width of all of them. This is not what we want, so for
                # those we create a fresh Layout object. The default min_width and width
                # are copied over so that they remain consistent with the other fields.
                widget.layout = Layout(
                    max_width=file_chooser_max_width,
                    min_width=widget.layout.min_width,
                    width=widget.layout.width,
                )
            continue

        if max_field_width is not None:
            widget.layout.max_width = max_field_width

    # The save and revert buttons should be enabled only when the user has made a
    # change AND the value in the widget is a valid pydantic model.
    # We begin by disabling the buttons.
    ui.savebuttonbar.bn_save.disabled = True
    ui.savebuttonbar.bn_revert.disabled = True

    # Now we add observers to enable/disable the buttons based on the validity of
    # the value and whether there are unsaved changes.

    # The save button should be enabled only when the user has made a change AND
    # the value in the widget is a valid pydantic model.
    ui.is_valid.observe(
        _handle_save_revert_button_state(ui, ui.savebuttonbar.bn_save), "value"
    )
    ui.savebuttonbar.observe(
        _handle_save_revert_button_state(ui, ui.savebuttonbar.bn_save),
        "unsaved_changes",
    )

    # The revert button should be enabled only when there are unsaved changes.
    ui.savebuttonbar.observe(
        _handle_save_revert_button_state(
            ui, ui.savebuttonbar.bn_revert, must_be_valid=False
        ),
        "unsaved_changes",
    )

    return ui


def _handle_save_revert_button_state(widget, button, must_be_valid=True):
    """
    Return a callback that will enable/disable the save and revert buttons based
    on the validity of the value and whether there are unsaved changes.

    Parameters
    ----------
    widget : `ipyautoui.AutoUi`
        The user interface widget.

    button : `ipywidgets.Button`
        The button to enable/disable based on the widget state.

    must_be_valid : bool, optional
        If `True`, the button will only be enabled if the value in the widget is
        a valid pydantic model. If `False`, the button will be enabled regardless
        of the validity of the value. Default is `True`.
    """

    def handler(_):
        """
        A handler must take an argument but we don't use it here.
        """
        valid_flag = widget.is_valid.value if must_be_valid else True
        needs_to_save = valid_flag and widget.savebuttonbar.unsaved_changes
        button.disabled = not needs_to_save

    return handler



import warnings

import numpy as np
from astropy.nddata import Cutout2D
from astropy.nddata.utils import NoOverlapError
from astropy.stats import sigma_clipped_stats
from astropy.utils import lazyproperty
from photutils.centroids import centroid_2dg, centroid_com
from photutils.profiles import CurveOfGrowth, RadialProfile

from stellarphot.photometry import calculate_noise

__all__ = ["find_center", "CenterAndProfile"]


def find_center(image, center_guess, cutout_size=30, max_iters=10, match_limit=3):
    """
    Find the centroid of a star from an initial guess of its position. Originally
    written to find star from a mouse click.

    Parameters
    ----------

    image : `astropy.nddata.CCDData` or numpy array
        Image containing the star.

    center_guess : array or tuple
        The position, in pixels, of the initial guess for the position of
        the star. The coordinates should be horizontal first, then vertical,
        i.e. opposite the usual Python convention for a numpy array.

    cutout_size : int, optional
        The default width of the cutout to use for finding the star.

    max_iters : int, optional
        Maximum number of iterations to go through in finding the center.

    match_limit : int, optional
        Maximum number of pixels to allow the COM centroid and Gaussian center to
        differ.

    Returns
    -------

    cen : array
        The position of the star, in pixels, as found by the centroiding
        algorithm.


    Raises
    ------
    RuntimeError
        If the centroiding algorithm fails to converge on a star eitehr because a cutout
        has only ``NaN`` values or because the centroiding algorithm fails to converge
        or because the centroid determined by ``centroid_com`` and ``centroid_2dg``
        differ by more than ``match_limit`` pixels.

    Notes
    -----
    This function tries to identify the centroid of a star in a small region around an
    image position. The original mtivation was to find the star near a mouse click on
    an image. The approach is to generate a cutout around the initial guess position,
    then use the centroid_com function from photutils to find the centroid of the star.
    A new cutout is then generated around the new centroid position, and the process
    is repeated until the centroid converges.

    Convergence is determined by three criteria:

    1. The centroid of the cutout must be within 3 pixels of the center of the cutout.
    2. The centroid of the cutout must be within 0.1 pixels of the previous centroid.
    3. The first two criteria must be met within the maximum number of iterations.

    If the first two criteria are satisfied then the centroid is found by fitting a
    Gaussian to the cutout. If the two centroids differ by more than match_limit pixels
    then an error is raised.
    """
    pad = cutout_size // 2
    x, y = center_guess

    # Keep track of iterations
    cnt = 0

    # Grab the cutout...
    sub_data = Cutout2D(image, center_guess, (cutout_size, cutout_size), mode="trim")
    # ...do stats on it...
    _, sub_med, _ = sigma_clipped_stats(sub_data.data)
    # ...and centroid.

    # Exclude negative pixels from initial centroid. If there is a dim star this helps
    # ensure the star ends up centered since pixels with negative values are likely
    # background.
    # See also Howell, Handbook of CCD Astronomy, 2nd ed., p. 105
    mask = (sub_data.data - sub_med) < 0
    x_cm, y_cm = centroid_com(sub_data.data - sub_med, mask=mask)

    # Translate centroid back to original image
    cen = np.array(sub_data.to_original_position((x_cm, y_cm)))

    # ceno is the "original" center guess, set it to something nonsensical here
    ceno = np.array([-100, -100])

    while cnt <= max_iters and (  # Iteration limit has not been reached
        np.abs(np.array([x_cm, y_cm]) - pad).max() > 3  # Centroid > 3 pix from center
        or np.abs(cen - ceno).max() > 0.1  # Centroid has not converged
    ):
        try:
            sub_data = Cutout2D(image, cen, (cutout_size, cutout_size), mode="trim")
        except NoOverlapError as err:
            raise RuntimeError(
                f"Centroid finding failed, previous was {ceno}, current is {cen}"
            ) from err
        _, sub_med, _ = sigma_clipped_stats(sub_data.data)

        mask = (sub_data.data - sub_med) < 0
        x_cm, y_cm = centroid_com(sub_data.data - sub_med, mask=mask)
        ceno = cen
        cen = np.array(sub_data.to_original_position((x_cm, y_cm)))
        if not np.all(~np.isnan(cen)):
            raise RuntimeError(
                f"Centroid finding failed, previous was {ceno}, current is {cen}"
            )
        cnt += 1

    # Is we hit the max number of iterations, raise an error
    if max_iters > 1 and cnt > max_iters:
        raise RuntimeError("Centroid finding did not converge")

    # Get the final centroid position by fitting a gaussian
    # We may not have converged on a star, so capture any warning about a
    # bad fit.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ceng = centroid_2dg(sub_data.data - sub_med)
    ceng = np.array(sub_data.to_original_position(ceng))

    # Confirm that we actually found a star by comparing two centroiding
    # methods. If we found a star they should be very close to each other.
    if np.linalg.norm(cen - ceng) > match_limit:
        raise RuntimeError(
            "Centroid did not converge on a star. "
            f"Got {cen} from centroid_com and {ceng} from centroid_2dg."
        )

    return ceng


class CenterAndProfile:
    """
    Class to determine center of and hold radial profile information for a star.

    Parameters
    ----------
    data : `astropy.nddata.CCDData` or numpy array
        Image data

    center_approx : list-like
        x, y position of the center in pixel coordinates, i.e. horizontal
        coordinate then vertical.

    cutout_size : int, optional
        Width of the rectangular image cutout to use in looking for a star.

    profile_radius : int, optional
        Maximum radius to use in constructing the profile.

    max_iters : int, optional
        Maximum number of iterations to go through in finding the center.

    match_limit : int, optional
        Maximum number of pixels to allow the COM centroid and Gaussian center to
        differ.
    """

    def __init__(
        self,
        data,
        center_approx,
        centering_cutout_size=30,
        profile_radius=None,
        max_iters=10,
        match_limit=3,
    ):
        # For centering we need a fairly small cutout containing only one star.
        self._cen = find_center(
            data,
            center_approx,
            cutout_size=centering_cutout_size,
            match_limit=match_limit,
            max_iters=max_iters,
        )
        self._data = data

        if profile_radius is None:
            profile_radius = centering_cutout_size // 2

        # The cutout for profiling needs to be at least twice the profile radius to
        # ensure that the profile is not truncated.
        self._profile_cutout = Cutout2D(
            data, self.center, (2 * profile_radius, 2 * profile_radius)
        )

        self._max_iters = max_iters

        radii = np.linspace(0, profile_radius, profile_radius + 1)
        # Get a rough profile with rough background subtraction -- note that
        # NO background subtraction does not work.
        background = sigma_clipped_stats(self.profile_cutout.data)[1]
        self._radial_profile = RadialProfile(
            self.profile_cutout.data - background, self.cutout_center, radii
        )

        self._sky_area = None

        # Do proper background subtraction for the final radial profile
        self._radial_profile = RadialProfile(
            self.profile_cutout.data - self.sky_pixel_value, self.cutout_center, radii
        )

    @property
    def center(self):
        """
        x, y position of the center of the star.
        """
        return self._cen

    @property
    def FWHM(self):
        """
        Full-width half-max of the radial profile.
        """
        return self.radial_profile.gaussian_fwhm

    @property
    def HWHM(self):
        """
        Half-width half-max of the radial profile.
        """
        return self.FWHM / 2

    @property
    def profile_cutout(self):
        """
        Cutout image around the star.
        """
        return self._profile_cutout

    @property
    def cutout_center(self):
        return self.profile_cutout.to_cutout_position(self.center)

    @property
    def radial_profile(self):
        """
        Radial profile of the star.
        """
        return self._radial_profile

    @property
    def max_iters(self):
        """
        Maximum number of iterations to go through in finding the center.
        """
        return self._max_iters

    @lazyproperty
    def normalized_profile(self):
        """
        Radial profile scaled to have a maximum of 1.
        """
        # This seems to be what photutils does under the hood
        return self.radial_profile.profile / self.radial_profile.profile.max()

    @lazyproperty
    def pixel_values_in_profile(self):
        """
        Pixel values in the radial profile.
        """
        radii = []
        pixel_values = []
        idx = 1
        cen_in_cutout = self.profile_cutout.to_cutout_position(self.center)
        for ap in self.radial_profile.apertures:
            idx += 1
            # Calculate the distance of each pixel from the center
            grid_x, grid_y = np.mgrid[
                : self.profile_cutout.shape[0], : self.profile_cutout.shape[1]
            ]
            dist_from_cen = np.sqrt(
                (grid_x - cen_in_cutout[1]) ** 2 + (grid_y - cen_in_cutout[0]) ** 2
            )

            # Get only the data in this aperture
            ap_mask = ap.to_mask(method="center")
            ap_data = ap_mask.multiply(self.profile_cutout.data)
            dist_from_cen = ap_mask.multiply(dist_from_cen)

            # Drop any data where the aperture mask was zero and flatten
            good_data = ap_data != 0
            ap_exact_radius = dist_from_cen[good_data].flatten()
            ap_data = ap_data[good_data].flatten()

            # Sort so that plots look ok
            sorted_radii = np.argsort(ap_exact_radius)
            radii.extend(ap_exact_radius[sorted_radii])
            pixel_values.extend(ap_data[sorted_radii])

        radii = np.array(radii)
        pixel_values = np.array(pixel_values)
        # Subtract the background
        return radii, pixel_values - self.sky_pixel_value

    @lazyproperty
    def curve_of_growth(self):
        """
        Curve of growth for the star.
        """
        self._cog = CurveOfGrowth(
            self._data - self.sky_pixel_value,
            self.center,
            self.radial_profile.radii + 1,
        )
        return self._cog

    @lazyproperty
    def sky_pixel_value(self):
        """
        Pixel values for the sky, i.e. more than 3 FWHM from the star.
        """
        grid_x, grid_y = np.mgrid[
            : self.profile_cutout.shape[0], : self.profile_cutout.shape[1]
        ]
        x_s, y_s = self.profile_cutout.to_cutout_position(self.center)
        dist_from_star = np.sqrt((grid_x - x_s) ** 2 + (grid_y - y_s) ** 2)
        mask = dist_from_star > self.FWHM * 3
        self._sky_area = mask.sum()
        _, median, _ = sigma_clipped_stats(self.profile_cutout.data[mask])
        return median

    @lazyproperty
    def sky_area(self):
        """
        Area of the sky annulus.
        """
        if self._sky_area is None:
            # sky area is set as a side effect of this....
            _ = self.sky_pixel_value

        return self._sky_area

    def noise(self, camera, exposure):
        """
        Noise in the star.

        Parameters
        ----------

        camera : `stellarphot.settings.Camera`
            Camera settings.

        exposure : float
            Exposure time in seconds.

        Returns
        -------
        noise : `numpy.ndarray` of float
            Noise calculated using the CCD equation.
        """
        return calculate_noise(
            camera,
            counts=self.curve_of_growth.profile,
            sky_per_pix=self.sky_pixel_value,
            aperture_area=self.curve_of_growth.area,
            annulus_area=0,
            exposure=exposure,
        )

    def snr(self, camera, exposure):
        """
        Signal to noise ratio of the star.

        Parameters
        ----------
        camera : `stellarphot.settings.Camera`
            Camera settings.

        exposure : float
            Exposure time in seconds.

        Returns
        -------
        snr : `numpy.ndarray` of float
            Signal to noise ratio.
        """
        return camera.gain * self.curve_of_growth.profile / self.noise(camera, exposure)



import warnings

import numpy as np
import pytest
from astropy import units as u
from astropy.stats import gaussian_sigma_to_fwhm
from astropy.table import QTable
from astropy.utils.exceptions import AstropyUserWarning

from stellarphot.photometry import compute_fwhm, source_detection
from stellarphot.photometry.tests.fake_image import FakeImage

# Make sure the tests are deterministic by using a random seed
SEED = 5432985


@pytest.mark.parametrize("units", [u.pixel, None])
def test_compute_fwhm(units):
    fake_image = FakeImage(seed=SEED)
    sources = fake_image.sources
    if units is not None:
        # It turns out having a unit on a column is not the same as
        # things in the column having units. The construct below ensures
        # that the source table values have units.
        # Do not try: sources['x_mean'] = sources['x_mean'] * units
        # Turns out individual values do NOT have units in that case.
        sources["x_mean"] = [v * units for v in sources["x_mean"]]
        sources["y_mean"] = [v * units for v in sources["y_mean"]]

    fwhm_x, fwhm_y = compute_fwhm(
        fake_image.image, sources, x_column="x_mean", y_column="y_mean"
    )

    expected_fwhm = np.array(sources["x_stddev"] * gaussian_sigma_to_fwhm)
    assert np.allclose(fwhm_x, expected_fwhm, rtol=1e-2)


def test_compute_fwhm_with_NaNs():
    # Regression test for https://github.com/feder-observatory/stellarphot/issues/161
    # We should be able to find FWHM for a source even with NaNs in the image.
    fake_image = FakeImage(seed=SEED)
    sources = fake_image.sources
    x, y = sources["x_mean"].astype(int)[0], sources["y_mean"].astype(int)[0]
    image = fake_image.image.copy()

    # Add a NaN to the image at the location of the first source. Note the
    # usual row/column swap when going to x/y coordinates.
    image[y, x] = np.nan

    # We expect a warning about NaNs in the image, so catch it
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="Non-Finite input data has been removed",
            category=AstropyUserWarning,
        )
        fwhm_x, fwhm_y = compute_fwhm(
            image, sources, x_column="x_mean", y_column="y_mean", fit=True
        )

    expected_fwhm = np.array(sources["x_stddev"] * gaussian_sigma_to_fwhm)
    assert np.allclose(fwhm_x, expected_fwhm, rtol=1e-2)


def test_detect_source_number_location():
    """
    Make sure we detect the sources in the input table....
    """
    fake_image = FakeImage(seed=SEED)
    sources = QTable(
        fake_image.sources,
        units={
            "x_mean": u.pixel,
            "y_mean": u.pixel,
            "x_stddev": u.pixel,
            "y_stddev": u.pixel,
        },
    )
    # print(sources)
    # Pass only one value for the sky background for source detection
    sky_per_pix = sources["sky_per_pix_avg"].mean()
    found_sources = source_detection(
        fake_image.image,
        fwhm=2 * sources["x_stddev"].mean(),
        threshold=10,
        sky_per_pix_avg=sky_per_pix,
    )
    # Sort by flux so we can reliably match them
    sources.sort("amplitude")
    found_sources.sort("flux")

    # Do we have the right number of sources?
    assert len(sources) == len(found_sources)

    for inp, out in zip(sources, found_sources, strict=True):
        # Do the positions match?
        np.testing.assert_allclose(out["xcenter"], inp["x_mean"], rtol=1e-5, atol=0.05)
        np.testing.assert_allclose(out["ycenter"], inp["y_mean"], rtol=1e-5, atol=0.05)
        np.testing.assert_allclose(
            gaussian_sigma_to_fwhm * (inp["x_stddev"] + inp["y_stddev"]) / 2,
            out["width"],
            rtol=1e-5,
            atol=0.05,
        )


def test_detect_source_with_padding():
    """
    Make sure we detect the sources in the input table....
    """
    fake_image = FakeImage(seed=SEED)
    sources = QTable(
        fake_image.sources,
        units={
            "x_mean": u.pixel,
            "y_mean": u.pixel,
            "x_stddev": u.pixel,
            "y_stddev": u.pixel,
        },
    )
    # Pass only one value for the sky background for source detection
    sky_per_pix = sources["sky_per_pix_avg"].mean()
    # Padding was chosen to be large enough to ensure that one of the sources in
    # test_sources.csv would land too close to the edge of the image.
    found_sources = source_detection(
        fake_image.image,
        fwhm=2 * sources["x_stddev"].mean(),
        threshold=10,
        sky_per_pix_avg=sky_per_pix,
        padding=95,
    )

    # Did we drop one source because it was too close to the edge?
    assert len(sources) - 1 == len(found_sources)



from pathlib import Path

import pytest
from astropy.utils.data import get_pkg_data_filename

from stellarphot.gui_tools import ComparisonAndSeeing


class TestComparisonAndSeeing:
    def copy_fits_to_temp(self, tmp_path):
        # Copy a fits file to a temporary directory
        fits_file = get_pkg_data_filename(
            "tests/data/wasp-10-tiny.fit.bz2", package="stellarphot"
        )
        fits_file = Path(fits_file)
        new_fits_file = tmp_path / fits_file.name
        new_fits_file.write_bytes(fits_file.read_bytes())
        return new_fits_file

    def test_init(self):
        # Just make sure we can create the object
        ComparisonAndSeeing()

    @pytest.mark.remote_data
    def test_setting_fits_file_in_comp_affects_seeing(self, tmp_path):
        # The only tricky-ish thing this class does is make sure that the
        # fits_file in the seeing widget and the fits_file in the comparison
        # widget are always the same. This test checks that setting the fits
        # file in the comparison widget updates the fits file in the seeing
        # widget.
        comp = ComparisonAndSeeing()
        file = self.copy_fits_to_temp(tmp_path)
        # Make sure the file exists -- necessary in some versions of ipyautoui
        file.touch()
        comp.comparison.fits_file.file_chooser._value = str(file)
        assert comp.seeing.fits_file.file_chooser._value == str(file)
        assert comp.comparison.fits_file.file_chooser.selected == str(file)

    @pytest.mark.remote_data
    def test_setting_fits_file_in_seeing_affects_comp(self, tmp_path):
        # This is the same as the previous test, but in the other direction.
        comp = ComparisonAndSeeing()
        file = self.copy_fits_to_temp(tmp_path)
        # Make sure the file exists -- necessary in some versions of ipyautoui
        file.touch()
        comp.seeing.fits_file.file_chooser._value = str(file)
        assert comp.comparison.fits_file.file_chooser._value == str(file)
        assert comp.seeing.fits_file.file_chooser.selected == str(file)



import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.modeling.models import Gaussian1D
from astropy.nddata import CCDData
from astropy.utils.data import get_pkg_data_filename
from photutils.datasets import make_noise_image

from stellarphot.photometry import CenterAndProfile, find_center
from stellarphot.photometry.tests.fake_image import make_gaussian_sources_image
from stellarphot.settings import Camera
from stellarphot.settings.tests.test_models import TEST_CAMERA_VALUES

SHAPE = (300, 300)
RANDOM_SEED = 1230971


class TestCenter:
    # The scope="function" below is not required (it is the default option) but
    # being explicit seems good, and we want the scope to be function so that each test
    # gets its own copy of the table.
    # Setting autouse=True means that this fixture will be run for every test in this
    # class. To get the data from the fixture we need to store it somewhere, so we store
    # it in self.profile_stars.
    @pytest.fixture(scope="function", autouse=True)
    def make_star_table(self, profile_stars):
        self.profile_stars = profile_stars

    def test_find_center_no_noise_good_guess(self):
        image = make_gaussian_sources_image(SHAPE, self.profile_stars)
        # Good initial guess, no noise, should converge in one try
        cen1 = find_center(image, (31, 41), max_iters=1)
        np.testing.assert_allclose(cen1, [30, 40])
        self.profile_stars = self.profile_stars[:-1]

    def test_find_center_noise_bad_guess(self):
        image = make_gaussian_sources_image(SHAPE, self.profile_stars)
        noise = make_noise_image(
            SHAPE, distribution="gaussian", mean=0, stddev=5, seed=RANDOM_SEED
        )
        cen2 = find_center(image + noise, [40, 50], max_iters=1)
        # Bad initial guess, noise, should take more than one try...
        with pytest.raises(AssertionError):
            np.testing.assert_allclose(cen2, [30, 40])

    def test_find_center_noise_good_guess(self):
        image = make_gaussian_sources_image(SHAPE, self.profile_stars)
        noise = make_noise_image(
            SHAPE, distribution="gaussian", mean=0, stddev=5, seed=RANDOM_SEED
        )
        # Trying again with several iterations should work
        cen3 = find_center(image + noise, [31, 41], max_iters=20)
        # Tolerance chosen based on some trial and error
        np.testing.assert_allclose(cen3, [30, 40], atol=0.02)

    def test_find_center_no_noise_star_at_edge(self):
        # Trying to put the star at the edge of the initial guess
        image = make_gaussian_sources_image(SHAPE, self.profile_stars)
        cen = find_center(image, [45, 65], max_iters=20)

        np.testing.assert_allclose(cen, [30, 40], atol=0.02)

    def test_find_center_no_star(self):
        # No star anywhere near the original guess
        image = make_gaussian_sources_image(SHAPE, self.profile_stars)
        # Offset the mean from zero to avoid nan center
        noise = make_noise_image(
            SHAPE, distribution="gaussian", mean=1000, stddev=5, seed=RANDOM_SEED
        )

        with pytest.raises(RuntimeError, match="Centroid did not converge on a star"):
            find_center(image + noise, [50, 200], max_iters=10)

    def test_find_center_dim_star(self):
        # Regression test for #352, in which a dim star is improperly centered.
        # The cutout loaded below is from an image of the field of WASP-10, and the star
        # in question has Gaia DR3 ID that is stored in the header. The Gaia position
        # is also stored in the header, and is what is taken to be the "correct"
        # position of the star.
        #
        # There is only one star in this cutout.
        #
        # For the record, the Gaia DR3 ID is 1909763087978475392.

        name = get_pkg_data_filename(data_name="data/center_cutout_1.fits")
        ccd = CCDData.read(name)

        true_coordinate = SkyCoord(ccd.header["gaia_coord"])
        true_pixel_center = ccd.wcs.world_to_pixel(true_coordinate)

        # At least with a faint star and large cutout, the
        # centroid as determined by COM and the centroid determined by the Gaussian fit
        # are about 20 pixels apart (pre-bug-fix).
        with pytest.raises(RuntimeError, match="Centroid did not converge on a star"):
            center = find_center(
                ccd.data,
                (40, 42),
                cutout_size=80,
                max_iters=10,
            )

        # Now try with a much smaller cutout size
        center = find_center(ccd.data, (40, 42), cutout_size=20, max_iters=10)
        # Check that we got a good center...
        assert np.linalg.norm(center - true_pixel_center) < 2


class TestRadialProfile:
    # The scope="function" below is not required (it is the default option) but
    # being explicit seems good, and we want the scope to be function so that each test
    # gets its own copy of the table.
    # Setting autouse=True means that this fixture will be run for every test in this
    # class. To get the data from the fixture we need to store it somewhere, so we store
    # it in self.profile_stars.
    @pytest.fixture(scope="function", autouse=True)
    def make_star_table(self, profile_stars):
        self.profile_stars = profile_stars

    def test_radial_profile(self):
        # Test that both curve of growth and radial profile are correct

        image = make_gaussian_sources_image(SHAPE, self.profile_stars)
        for row in self.profile_stars:
            cen = find_center(image, (row["x_mean"], row["y_mean"]), max_iters=10)

            # The "stars" have FWHM around 9.5, so make the cutouts used for finding the
            # stars fairly big -- the bare minimum would be a radius of 3 FWHM, which is
            # a cutout size around 60.
            rad_prof = CenterAndProfile(
                image, cen, centering_cutout_size=60, profile_radius=30
            )

            # Test that the curve of growth is correct

            # Numerical value below is integral of input 2D gaussian, 2pi A sigma^2
            expected_integral = 2 * np.pi * row["amplitude"] * row["x_stddev"] ** 2
            np.testing.assert_allclose(
                rad_prof.curve_of_growth.profile[-1], expected_integral, atol=50
            )

            # Test that the radial profile is correct by comparing pixel values to a
            # gaussian fit to the profile.
            data_radii, data_counts = rad_prof.pixel_values_in_profile
            expected_profile = rad_prof.radial_profile.gaussian_fit(data_radii)

            np.testing.assert_allclose(data_counts, expected_profile, atol=20)

    def test_radial_profile_exposure_is_nan(self):
        # Check that using an exposure value of NaN returns NaN for the SNR and noise
        image = make_gaussian_sources_image(SHAPE, self.profile_stars)

        cen = find_center(image, (50, 50), max_iters=10)

        rad_prof = CenterAndProfile(
            image, cen, centering_cutout_size=60, profile_radius=30
        )

        c = Camera(**TEST_CAMERA_VALUES)
        assert np.isnan(rad_prof.snr(c, np.nan)[-1])
        assert np.isnan(rad_prof.noise(c, np.nan)[-1])
        assert all(np.isfinite(rad_prof.curve_of_growth.profile))
        assert all(np.isfinite(rad_prof.radial_profile.profile))

    def test_radial_profile_with_background(self):
        # Regression test for #328 -- image with a background level
        image = make_gaussian_sources_image(SHAPE, self.profile_stars)
        noise_stdev = 10
        image = image + make_noise_image(
            image.shape,
            distribution="gaussian",
            mean=100,
            stddev=noise_stdev,
            seed=43917,
        )
        for row in self.profile_stars:
            cen = find_center(image, (row["x_mean"], row["y_mean"]), max_iters=10)

            # The "stars" have FWHM around 9.5, so make the cutouts used for finding the
            # stars fairly big -- the bare minimum would be a radius of 3 FWHM, which is
            # a cutout size around 60.
            rad_prof = CenterAndProfile(
                image, cen, centering_cutout_size=60, profile_radius=30
            )

            # Numerical value below is integral of input 2D gaussian, 2pi A sigma^2
            expected_integral = 2 * np.pi * row["amplitude"] * row["x_stddev"] ** 2

            # The standard deviation in the sum of N gaussian random variables with
            # standard deviation SD is
            #    σ = sqrt(N × SD^2)
            # The curve of growth includes the sum of a bunch of pixels which each have
            # a standard deviation of 10, so the standard deviation of the sum of those
            # pixels is given by the formula above, with N being the number of pixels
            # in the curve.

            expected_stddev = np.sqrt(
                rad_prof.curve_of_growth.area[-1] * noise_stdev**2
            )
            print(
                expected_stddev,
                rad_prof.curve_of_growth.profile[-1] - expected_integral,
            )

            # With the seed above the difference is just under 1.5 standard deviations.
            np.testing.assert_allclose(
                rad_prof.curve_of_growth.profile[-1],
                expected_integral,
                atol=1.5 * expected_stddev,
            )

            # Test that the radial profile is correct by comparing pixel values to a
            # gaussian fit to the profile.
            data_radii, data_counts = rad_prof.pixel_values_in_profile
            expected_profile = rad_prof.radial_profile.gaussian_fit(data_radii)

            # The test here is that the difference between the actual profile and the
            # expected is itself a Gaussian distribution with standard deviation very
            # roughly equal to the standard deviation of the noise we put in.
            differences = data_counts - expected_profile
            counts, bin_edges = np.histogram(differences)
            bin_centers = (bin_edges[1:] + bin_edges[:-1]) / 2
            g1d_init = Gaussian1D()
            fitter = LevMarLSQFitter()
            g1d = fitter(g1d_init, bin_centers, counts)

            assert np.abs(g1d.stddev.value - noise_stdev) < 1.5
            assert np.abs(g1d.mean.value) < 1

    def test_radial_profile_bigger_profile_than_cutout(self):
        # Test that the cutout, used for finding the star, can be smaller than
        # the profile radius.
        image = make_gaussian_sources_image(SHAPE, self.profile_stars)

        # Just look at one star in this test, the last one, which is far from the edges.
        # Prior to a change in CenterAndProfile, this would have raised an error because
        # the same cutout size was used for the profile and the centering. The result if
        # the profile size was larger was that the profile eventually had only NaNs in
        # the outermost annuli.
        profile = CenterAndProfile(
            image,
            (self.profile_stars["x_mean"][-1], self.profile_stars["y_mean"][-1]),
            centering_cutout_size=20,
            profile_radius=50,
        )

        assert profile.profile_cutout.shape == (100, 100)

    def test_radial_profile_no_profile_size(self):
        # Test that when we do not provide a profile size it is half the cutout size
        image = make_gaussian_sources_image(SHAPE, self.profile_stars)
        profile = CenterAndProfile(
            image,
            (self.profile_stars["x_mean"][-1], self.profile_stars["y_mean"][-1]),
            centering_cutout_size=50,
        )

        # Last point should be the average of the last two bin edges, which is 24.5
        assert profile.radial_profile.radius[-1] == 24.5



import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.time import Time

from stellarphot import PhotometryData
from stellarphot.differential_photometry.aij_rel_fluxes import calc_aij_relative_flux


def _repeat(array, count):
    return np.concatenate([array for _ in range(count)])


def _raw_photometry_table():
    """
    Generate an input raw photometry table and expected flux ratios
    for use in tests.
    """

    n_times = 10
    n_stars = 4
    # How about ten times...
    times = Time("2018-06-25T01:00:00", format="isot", scale="utc")
    times = times + np.arange(n_times) * 30 * u.second
    times = times.value

    # and four stars
    star_ra = 250.0 * u.degree + np.arange(n_stars) * 10 * u.arcmin
    star_dec = np.array([45.0] * n_stars) * u.degree
    fluxes = np.array([10000.0, 20000, 30000, 40000]) * u.adu
    errors = (np.sqrt(fluxes.value) + 50) * u.electron
    star_ids = np.arange(1, 5, dtype="int")

    # Stars 2, 3 and 4 will be the comparison stars
    comp_stars = np.array([0, 1, 1, 1])
    expected_comp_fluxes = np.sum(fluxes[1:])

    comp_flux_offset = -comp_stars * fluxes
    expected_flux_ratios = fluxes / (expected_comp_fluxes + comp_flux_offset)

    comp_error_total = np.sqrt((errors[1:] ** 2).sum())

    expected_flux_error = (
        fluxes
        / expected_comp_fluxes
        * np.sqrt(errors**2 / fluxes**2 + comp_error_total**2 / expected_comp_fluxes**2)
    )

    raw_table = Table(
        data=[
            np.sort(_repeat(times, n_stars)),
            _repeat(star_ra, n_times),
            _repeat(star_dec, n_times),
            _repeat(fluxes, n_times),
            _repeat(errors, n_times),
            _repeat(star_ids, n_times),
        ],
        names=[
            "date-obs",
            "ra",
            "dec",
            "aperture_net_cnts",
            "noise_electrons",
            "star_id",
        ],
        units=[
            None,
            u.degree,
            u.degree,
            u.adu,
            u.electron,
            None,
        ],
    )

    photom = PhotometryData(raw_table)
    # MAKE SURE to return photom, not raw_table, below to trigger the bug
    # https://github.com/feder-observatory/stellarphot/issues/421
    # in which, it turns out, QTable columns with units cannot be aggregated.
    return expected_flux_ratios, expected_flux_error, photom, photom[1:4]


@pytest.mark.parametrize("comp_ra_dec_have_units", [True, False])
@pytest.mark.parametrize("star_ra_dec_have_units", [True, False])
@pytest.mark.parametrize("in_place", [True, False])
def test_relative_flux_calculation(
    in_place, star_ra_dec_have_units, comp_ra_dec_have_units
):
    # In addition to checking the flux calculation values, this is also a regression
    # test for #421.
    expected_flux, expected_error, input_table, comp_star = _raw_photometry_table()

    # Try doing it all at once
    n_times = len(np.unique(input_table["date-obs"]))
    all_expected_flux = _repeat(expected_flux, n_times)
    all_expected_error = _repeat(expected_error, n_times)

    if not star_ra_dec_have_units:
        input_table["ra"] = input_table["ra"].data
        input_table["dec"] = input_table["dec"].data

    if not comp_ra_dec_have_units:
        comp_star["ra"] = comp_star["ra"].data
        comp_star["dec"] = comp_star["dec"].data

    output_table = calc_aij_relative_flux(input_table, comp_star, in_place=in_place)
    output_flux = output_table["relative_flux"]
    output_error = output_table["relative_flux_error"]

    np.testing.assert_allclose(output_flux, all_expected_flux)
    np.testing.assert_allclose(output_error, all_expected_error)
    if in_place:
        assert "relative_flux" in input_table.colnames
    else:
        assert "relative_flux" not in input_table.colnames


@pytest.mark.parametrize("bad_thing", ["RA", "NaN"])
def test_bad_comp_star(bad_thing):
    expected_flux, expected_error, input_table, comp_star = _raw_photometry_table()
    # We'll do modify the "bad" property for the last star in the last
    # image.

    # First, let's sort so the row we want to modify is the last one
    input_table.sort(["date-obs", "star_id"])

    # Force a copy of this row so we have access to the original values
    last_one = Table(input_table[-1])

    if bad_thing == "RA":
        # "Jiggle" one of the stars by moving it by a few arcsec in one image.
        coord_inp = SkyCoord(
            ra=last_one["ra"][0], dec=last_one["dec"][0], unit=u.degree
        )
        coord_bad_ra = coord_inp.ra + 3 * u.arcsecond
        input_table["ra"][-1] = coord_bad_ra
    elif bad_thing == "NaN":
        input_table["aperture_net_cnts"][-1] = np.nan

    output_table = calc_aij_relative_flux(input_table, comp_star, in_place=False)

    old_total_flux = comp_star["aperture_net_cnts"].sum()
    new_flux = old_total_flux - last_one["aperture_net_cnts"]
    # This works for target stars, i.e. those never in comparison set
    new_expected_flux = old_total_flux / new_flux * expected_flux

    # Oh wow, this is terrible....
    # Need to manually calculate for the only two that are still in comparison
    new_expected_flux[1] = (
        comp_star["aperture_net_cnts"][0] / comp_star["aperture_net_cnts"][1]
    )
    new_expected_flux[2] = (
        comp_star["aperture_net_cnts"][1] / comp_star["aperture_net_cnts"][0]
    )

    new_expected_flux[3] = expected_flux[3]
    if bad_thing == "NaN":
        new_expected_flux[3] = np.nan

    np.testing.assert_allclose(new_expected_flux, output_table["relative_flux"][-4:])



import os
from pathlib import Path

import ipywidgets as ipw
import pytest
from astropy.io import fits

from stellarphot.settings import (
    PhotometrySettings,
    PhotometryWorkingDirSettings,
    settings_files,
)
from stellarphot.settings.custom_widgets import PhotometryRunner
from stellarphot.settings.tests.test_models import TEST_PHOTOMETRY_SETTINGS


# See test_settings_file.TestSavedSettings for a detailed description of what the
# following fixture does. In brief, it patches the settings_files.PlatformDirs class
# so that the user_data_dir method returns the temporary directory.
@pytest.fixture(autouse=True)
def fake_settings_dir(mocker, tmp_path):
    mocker.patch.object(
        settings_files.PlatformDirs, "user_data_dir", tmp_path / "stellarphot"
    )


class TestPhotometryRunner:
    # This auto-used fixture changes the working directory to the temporary directory
    # and then changes back to the original directory after the test is done.
    @pytest.fixture(autouse=True)
    def change_to_tmp_dir(self, tmp_path):
        original_dir = os.getcwd()
        os.chdir(tmp_path)
        # Yielding here is important. It means that when the test is done, the remainder
        # of the function will be executed. This is important because the test is run in
        # a temporary directory and we want to change back to the original directory
        # when the test is done.
        yield
        os.chdir(original_dir)

    def test_photometry_runner_creation(self):
        # This test simply makes sure we can create the object
        photometry_runner = PhotometryRunner()
        assert isinstance(photometry_runner, ipw.Box)

    @pytest.mark.parametrize("do_photometry", [True, False])
    def test_photometry_runner_with_valid_settings(self, do_photometry):
        # This test makes sure that the PhotometryRunner widget can be used
        # to *start* a photometry run. It does not check the actual photometry
        # or the results of the photometry run, except to ensure that the expected
        # files are created.
        # Make a settings
        phot_settings = PhotometrySettings.model_validate(TEST_PHOTOMETRY_SETTINGS)

        # Save a copy to the working directory
        wd_settings = PhotometryWorkingDirSettings()
        wd_settings.save(phot_settings)

        # Make sure a source list file exists
        sources = Path(phot_settings.source_location_settings.source_list_file)
        sources.touch()

        # Make sure a fake fits file exists
        fake_fits = Path("fake.fits")
        fits_data = fits.PrimaryHDU(data=[[1, 2], [3, 4]])
        object_name = "Fake Object"
        fits_data.header["object"] = object_name
        fits_data.writeto(fake_fits)

        photometry_runner = PhotometryRunner()
        # Select a fits file
        photometry_runner.fitsopen.file_chooser.reset(".", fake_fits.name)
        photometry_runner.fitsopen.file_chooser._apply_selection()
        photometry_runner.fitsopen.file_chooser.value = fake_fits

        # Check that the message in the information box is as expected
        assert "Photometry will be performed" in photometry_runner.info_box.value
        assert object_name in photometry_runner.info_box.value

        if do_photometry:
            # Run the photometry
            photometry_runner.confirm._yes.click()
            assert "Photometry is running" in photometry_runner.info_box.value

            # Check that the expected file was created, in this case just the notebook.
            # No photometry is actually done because the source is empty and the "image"
            # has no stars.
            assert Path(photometry_runner.photometry_notebook_name).exists()

            # Make sure we do get the error we expect, which is generated when
            # the source list is read. It is an empty file, not actually an ecsv.
            with open(photometry_runner.photometry_notebook_name) as f:
                notebook_text = f.read()
                assert "InconsistentTableError" in notebook_text
        else:
            # Cancel the photometry
            photometry_runner.confirm._no.click()

            assert photometry_runner.info_box.value == ""



import json
from typing import Annotated

import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy.units import Quantity, Unit, get_physical_type
from pydantic import BaseModel, ValidationError

from stellarphot.settings.astropy_pydantic import (
    AstropyValidator,
    EquivalentTo,
    QuantityType,
    UnitType,
    WithPhysicalType,
    serialize_astropy_type,
)


class _UnitModel(BaseModel):
    # Dummy class for testing
    unit: UnitType


class _QuantityModel(BaseModel):
    # Dummy class for testing
    quantity: QuantityType


@pytest.mark.parametrize(
    "init",
    [
        Unit("m"),
        1,
        4.0,
        "meter",
        "parsec / fortnight",
        "",
    ],
)
def test_unit_initialization(init):
    # Make sure we can initialize from each of the ways a Unit can
    # be initialized
    expected = Unit(init)
    unit = _UnitModel(unit=init)
    assert expected == unit.unit


@pytest.mark.parametrize(
    "init",
    [
        -2 * Unit("m"),
        -2.0 * Unit("m"),
        1,
        1.0,
        "5 meter",
        "13 parsec / fortnight",
        "42",
    ],
)
def test_quantity_intialization(init):
    # Make sure we can initialize from each of the ways a Quantity can
    # be initialized
    expected = Quantity(init)
    quantity = _QuantityModel(quantity=init)
    assert expected == quantity.quantity


class _ModelEquivalentTo(BaseModel):
    unit_meter: Annotated[UnitType, EquivalentTo(equivalent_unit="m")]
    quantity_meter: Annotated[QuantityType, EquivalentTo(equivalent_unit="m")]


class _ModelWithPhysicalType(BaseModel):
    quant_physical_length: Annotated[QuantityType, WithPhysicalType("length")]
    unit_physical_time: Annotated[UnitType, WithPhysicalType("time")]


def test_equivalent_to():
    # Make sure we can annotate with an equivalent unit

    # This should succeed
    model = _ModelEquivalentTo(
        unit_meter="km",
        quantity_meter=Quantity(1, "mm"),
    )
    assert model.unit_meter == Unit("km")
    assert model.quantity_meter.to("m").value == model.quantity_meter.value * 1e-3

    # Now some failures

    with pytest.raises(ValidationError, match="Unit s is not equivalent to"):
        _ModelEquivalentTo(unit_meter="km", quantity_meter=Quantity("1 s"))

    with pytest.raises(ValidationError, match="Unit s is not equivalent to"):
        _ModelEquivalentTo(unit_meter="s", quantity_meter=Quantity("1 m"))


def test_equivalent_to_can_be_used_in_union():
    class ModelWithUnion(BaseModel):
        may_be_none: Annotated[QuantityType, EquivalentTo("second")] | None

    model = ModelWithUnion(may_be_none=None)
    assert model.may_be_none is None
    model = ModelWithUnion(may_be_none=Quantity(1, "s"))
    assert model.may_be_none == Quantity(1, "s")


def test_physical_type_can_be_used_in_union():
    class ModelWithUnion(BaseModel):
        may_be_none: Annotated[QuantityType, WithPhysicalType("time")] | None

    model = ModelWithUnion(may_be_none=None)
    assert model.may_be_none is None
    model = ModelWithUnion(may_be_none=Quantity(1, "s"))
    assert model.may_be_none == Quantity(1, "s")


def test_with_physical_type():
    # Make sure we can annotate with a physical type
    model = _ModelWithPhysicalType(
        quant_physical_length=Quantity(1, "m"),
        unit_physical_time=17 * Unit("second"),
    )

    assert get_physical_type(model.quant_physical_length) == "length"
    assert get_physical_type(model.unit_physical_time) == "time"

    # Now some failures
    # Pass a time in for physical type of length
    with pytest.raises(
        ValidationError, match="Unit of 1.0 s is not equivalent to length"
    ):
        _ModelWithPhysicalType(
            quant_physical_length=Quantity(1, "s"),
            unit_physical_time=Unit("second"),
        )

    # Pass a length in for physical type of time
    with pytest.raises(ValidationError, match="Unit of m is not equivalent to time"):
        _ModelWithPhysicalType(
            quant_physical_length=Quantity(1, "m"),
            unit_physical_time=Unit("meter"),
        )


def test_bad_physical_type_raises_error():
    with pytest.raises(ValueError, match="'not_a_type' is not a known physical type"):

        class Model(BaseModel):
            not_a_type: Annotated[QuantityType, WithPhysicalType("not_a_type")]


def test_quantity_type_with_invalid_quantity():
    with pytest.raises(ValidationError, match="It does not start with a number"):
        _QuantityModel(quantity="meter")


@pytest.mark.parametrize(
    "input_json_string",
    [
        '{"quantity": "1 m"}',
        '{"quantity": "1"}',
        '{"quantity": "3 second"}',
    ],
)
def test_initialize_quantity_with_json(input_json_string):
    # Make sure we can initialize a Quantity from a json string
    # where the quantity value is stored in the json as a string.
    model = _QuantityModel.model_validate_json(input_json_string)
    model_json = json.loads(model.model_dump_json())
    input_json = json.loads(input_json_string)

    assert Quantity(model_json["quantity"]) == Quantity(input_json["quantity"])


def test_initialize_quantity_with_json_invalid():
    # Make sure we get an error when the json string is has a value
    # that is a float (same fail happens for integer).
    # Since our json validation assumes the value is a string, this
    # should fail.
    with pytest.raises(ValidationError, match="Input should be a valid string"):
        _QuantityModel.model_validate_json('{"quantity": 14.0}')


@pytest.mark.parametrize(
    "input_json_string",
    [
        '{"unit": "m"}',
        '{"unit": "1"}',
        '{"unit": "parsec / fortnight"}',
    ],
)
def test_initialize_unit_with_json(input_json_string):
    # Make sure we can initialize a Unit from a json string
    # where the quantity value is stored in the json as a string.
    model = _UnitModel.model_validate_json(input_json_string)
    model_json = json.loads(model.model_dump_json())
    input_json = json.loads(input_json_string)

    assert Unit(model_json["unit"]) == Unit(input_json["unit"])


def test_initialize_unit_with_json_invalid():
    # Make sure we get an error when the json string is has a value
    # that is a float (same fail happens for integer).
    # Since our json validation assumes the value is a string, this
    # should fail.
    with pytest.raises(ValidationError, match="Input should be a valid string"):
        _UnitModel.model_validate_json('{"unit": 14.0}')


@pytest.mark.parametrize(
    "klass,input",
    [
        (Time, "2021-01-01T00:00:00"),
        (SkyCoord, "00h42m44.3s +41d16m9s"),
    ],
)
def test_time_quant_pydantic(klass, input):
    class Model(BaseModel):
        value: Annotated[klass, AstropyValidator]

    val = klass(input)
    model = Model(value=val)

    # Value should be correct
    assert model.value == val

    # model dump should fully serialize to standard python types
    assert model.model_dump()["value"] == serialize_astropy_type(val)

    # We should be able to create a new model from the dumped json...
    model2 = Model.model_validate_json(model.model_dump_json())

    if klass is SkyCoord:
        np.testing.assert_almost_equal(
            model2.value.separation(model.value).arcsec,
            0,
            decimal=10,
        )
    else:
        assert model2.value == model.value


def test_time_pydantic_invalid_value():
    class Model(BaseModel):
        value: Annotated[Time, AstropyValidator]

    with pytest.raises(ValidationError, match="Input should be an instance of Time"):
        Model(value="not a time")



from dataclasses import dataclass
from typing import Annotated, Any

from astropy.units import (
    PhysicalType,
    Quantity,
    Unit,
    UnitBase,
    UnitConversionError,
    get_physical_type,
)
from pydantic import (
    GetCoreSchemaHandler,
)
from pydantic_core import core_schema

__all__ = [
    "UnitType",
    "QuantityType",
    "EquivalentTo",
    "WithPhysicalType",
    "AstropyValidator",
]

_PHYSICAL_TYPES_URL = "https://docs.astropy.org/en/stable/units/ref_api.html#module-astropy.units.physical"

# Approach to validation of units was inspired by the GammaPy project
# which did it before we did:
# https://docs.gammapy.org/dev/_modules/gammapy/analysis/config.html

# Update for pydantic 2.0, based on the pydantic docs:
#   https://docs.pydantic.dev/latest/concepts/types/#handling-third-party-types


# This function will end up creating the instance from the input value.
def validate_by_instantiating(source_type):
    """
    Return a function that tries to create an instance of source_type from a value.
    The intended use of this is as a validator in pydantic.

    Parameters
    ----------

    source_type : Any
        The type to create an instance of.

    Returns
    -------
    function
        A function that tries to create an instance of source_type from a value.
    """

    def _validator(value):
        # If the value is valid we will be able to create an instance of the
        # source_type from it. For example, if source_type is astropy.units.Unit,
        # then we should be able to create a Unit from the value.
        try:
            result = source_type(value)
        except TypeError as err:
            # Need to raise a ValueError for pydantic to catch it as a
            # validation error.
            raise ValueError(str(err)) from err
        return result

    return _validator


class _UnitQuantTypePydanticAnnotation:
    """
    This class is used to annotate fields where validation consists of checking
    whether an instance can be created.

    In astropy, this includes `astropy.units.Unit` and `astropy.units.Quantity`.

    We return a pydantic_core.CoreSchema that behaves in the following ways:
        * When initializing from python:
            * A Unit or a Quantity will pass validation and be returned as-is
            * A string or a float will be used to create a Unit or a Quantity.
            * Nothing else will pass validation
        * When initializing from json:
            * The Unit or Quantity must be represented as a string in the json.
        * Serialization will always represent the Unit or Quantity as a string.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """
        We return a pydantic_core.CoreSchema that behaves in the following ways:

        * A Unit or a Quantity will pass validation and be returned as-is
        * A string or a float will be used to create a Unit or a Quantity.
        * Nothing else will pass validation
        * Serialization will always return just a string
        """
        # Both Unit and Quantity can be created from a string or a float or an
        # instance of the same type. So we need to check for all of those cases.

        # core_schema.chain_schema runs the value through each of the schema
        # in the list, in order. The output of one schema is the input to the next.

        # The schema below validates a Unit or Quantity from a string. The first link
        # in the chain is a schema that validates a string. The second link is a
        # schema that takes that string value and creates a Unit or Quantity from it.

        # This schema is used in three places by pydantic:
        #   1. When validating a python value
        #   2. When validating a json value
        #   3. When constructing a JSON schema for the model
        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(
                    validate_by_instantiating(source_type)
                ),
            ]
        )

        # This schema validates a Unit or Quantity from a float. The first link in the
        # chain is a schema that validates the input as a number. The second link is a
        # schema that takes that numeric value and creates a Unit or Quantity from it.

        # This schema is used in just one place by pydantic:
        #   1. When validating a python value
        from_float_schema = core_schema.chain_schema(
            [
                core_schema.float_schema(),
                core_schema.no_info_plain_validator_function(
                    validate_by_instantiating(source_type)
                ),
            ]
        )

        return core_schema.json_or_python_schema(
            # The next line specifies two things:
            #   1. The schema used to validate value from JSON. Since we are using the
            #      schema for a string value, the values in the JSON file must be
            #      strings, even though something like "{quantity: 1}" i.e. initializing
            #      from a number, would work in python. The reason for this choice is
            #      that the serialization is to a string, so that is what we expect from
            #      JSON. If we wanted to allow initialization from a number in JSON, we
            #      would need to use a union schema that consisted of
            #      from_str_schema and from_float_schema.
            #   2. The schema used to construct a JSON schema for the model. When you do
            #      `model_json_schema` with a `chain_schema`, then the first entry of
            #      the chain is used if `mode="validation"` and the last entry of the
            #      chain is used if `mode="serialization"`. I guess this makes sense,
            #      since the first thing in the chain has to handle the value coming
            #      from json,while the last thing generates the python value for the
            #      input. With the choice below we will *always* want`mode="validation"`
            #      because pydantic cannot generate a schema for a Unit or Quantity.
            json_schema=from_str_schema,
            # union_schema takes a list of schemas and returns a schema that
            # is the "best" match. See the link below for a description of
            # what counts as "best":
            #   https://docs.pydantic.dev/dev/concepts/unions/#smart-mode
            #
            # In short, schemas are tried from left-to-right, and an exact type match
            # wins.
            #
            # The construction below tries to make a value starting from a Unit,
            # a Quantity, a string, or a float. The first two are instances, so we
            # use is_instance_schema.
            python_schema=core_schema.union_schema(
                [
                    # Check if it's an instance first before doing any further work.
                    # Would be nice to provide a list of classes here instead
                    # of one-by-one.
                    core_schema.is_instance_schema(UnitBase),
                    core_schema.is_instance_schema(Quantity),
                    from_str_schema,
                    from_float_schema,
                ]
            ),
            # Serialize by converting the Unit or Quantity to a string.
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: str(instance)
            ),
        )


@dataclass
class EquivalentTo:
    """
    This can be used to annotate pydantic fields where the value should have a unit
    that is convertible to a given unit.

    Parameters
    ----------
    equivalent_unit : `astropy.units.Unit`
        The unit that the annotated field should be equivalent to.

    Examples
    --------
    >>> from typing import Annotated
    >>> from pydantic import BaseModel, ValidationError
    >>> from stellarphot.settings.astropy_pydantic import (
    ...     WithPhysicalType,
    ...     UnitType,
    ...     QuantityType
    ... )
    >>> class UnitModel(BaseModel):
    ...     length: Annotated[UnitType, EquivalentTo("m")]
    >>> UnitModel(length="lightyear")
    UnitModel(length=Unit("lyr"))
    >>> try:
    ...     UnitModel(length="second")
    ... except ValidationError as err:
    ...     print(err)
    1 validation error for UnitModel
    length
      Value error, Unit s is not equivalent to m...
    >>> # Next let's do a Quantity
    >>> class QuantityModel(BaseModel):
    ...     velocity: Annotated[QuantityType, EquivalentTo("m / s")]
    >>> QuantityModel(velocity="3 lightyear / year")
    QuantityModel(velocity=<Quantity 3. lyr / yr>)
    >>> try:
    ...     QuantityModel(velocity="3 parsec / lightyear")
    ... except ValidationError as err:
    ...     print(err)
    1 validation error for QuantityModel
    velocity
      Value error, Unit pc / lyr is not equivalent to m / s...
    """

    equivalent_unit: Unit
    """Unit that the annotated field should be equivalent to."""

    def __get_pydantic_core_schema__(
        self, source_type: Any, handler: GetCoreSchemaHandler
    ):
        def check_equivalent(value):
            # We are getting either a Unit or a Quantity. If it's a Quantity, we
            # need to get the unit from it.
            if isinstance(value, UnitBase):
                value_unit = value
            else:
                value_unit = value.unit

            try:
                value.to(self.equivalent_unit)
            except UnitConversionError:
                # Raise a ValueError for pydantic to catch it as a validation error.
                raise ValueError(
                    f"Unit {value_unit} is not equivalent to {self.equivalent_unit}"
                ) from None
            return value

        # Calling handler(source_type) will pass the result of this annotation
        # to the next annotation in the chain.
        return core_schema.no_info_after_validator_function(
            check_equivalent, handler(source_type)
        )

    def __hash__(self):
        return hash(self.equivalent_unit)


@dataclass
class WithPhysicalType:
    """
    This class can be used to annotate pydantic fields whose value
    should be of a specific physical type.

    Parameters
    ----------
    physical_type : str or `astropy.units.physical.PhysicalType`
        The physical type of the annotated field.

    Examples
    --------
    >>> from typing import Annotated
    >>> from pydantic import BaseModel, ValidationError
    >>> from stellarphot.settings.astropy_pydantic import (
    ...     WithPhysicalType,
    ...     UnitType,
    ...     QuantityType
    ... )
    >>> class UnitModel(BaseModel):
    ...     length: Annotated[UnitType, WithPhysicalType("length")]
    >>> UnitModel(length="meter")
    UnitModel(length=Unit("m"))
    >>> try:
    ...     UnitModel(length="second")
    ... except ValidationError as err:
    ...     print(err)
    1 validation error for UnitModel
    length
      Value error, Unit of s is not equivalent to length...
    >>> # Next let's do a Quantity
    >>> class QuantityModel(BaseModel):
    ...     velocity: Annotated[QuantityType, WithPhysicalType("speed")]
    >>> QuantityModel(velocity="3 meter / second")
    QuantityModel(velocity=<Quantity 3. m / s>)
    >>> try:
    ...     QuantityModel(velocity="3 meter")
    ... except ValidationError as err:
    ...     print(err)
    1 validation error for QuantityModel
    velocity
      Value error, Unit of 3.0 m is not equivalent to speed...
    """

    physical_type: str | PhysicalType

    def __post_init__(self):
        try:
            get_physical_type(self.physical_type)
        except ValueError as err:
            # Add a link to the astropy documentation for physical types
            # to the error message.
            raise ValueError(
                str(err)
                + f"\nSee {_PHYSICAL_TYPES_URL} for a list of valid physical types."
            ) from err

    def __get_pydantic_core_schema__(
        self, source_type: Any, handler: GetCoreSchemaHandler
    ):
        def check_physical_type(value):
            is_same = get_physical_type(value) == get_physical_type(self.physical_type)
            if is_same:
                return value
            else:
                raise ValueError(
                    f"Unit of {value} is not equivalent to {self.physical_type}"
                ) from None

        # As in the EquivalentTo annotation, calling handler(source_type) will pass
        # the result of this annotation to the next annotation in the chain.
        return core_schema.no_info_after_validator_function(
            check_physical_type, handler(source_type)
        )

    def __hash__(self):
        return hash(self.physical_type)


# We have lost default titles and examples, but that is maybe not so bad

# This is really nice compared to pydantic v1...
UnitType = Annotated[Unit, _UnitQuantTypePydanticAnnotation]

# Quantity type is really clean too
QuantityType = Annotated[Quantity, _UnitQuantTypePydanticAnnotation]


# It turns out all (almost all?) astropy types have, buried in them, a representation
# that is a python dictionary info._represent_as_dict and info._construct_from_dict.
# This is what is used to represent astropy
# objects in Tables and FITS files. So we can use this to create a json schema
# for the astropy types.
def serialize_astropy_type(value):
    """
    Serialize astropy objects like Time and SkyCoord to a dictionary.

    In principle, we ought to be able to use the astropy serialization stuff
    that is used in writing tables to ecsv here, but that is not quite working
    as of January 2024, so for now we fall back to a string representation when
    the value is a Unit or Quantity.

    Parameters
    ----------
    value : Any
        The value to serialize.

    Returns
    -------
    dict
        A dictionary representation of the astropy object.
    """

    def dict_rep(instance):
        """
        This is in a function so it only shows up in one place in the event the
        private astropy API changes.
        """
        return instance.info._represent_as_dict()

    # The if statement below is a bit of a hack. It's not clear to me how to
    # use the _represent_as_dict stuff to serialize something like a SkyCoord that
    # has nested astropy objects in it. So for now, we just return the string
    # representation of the objects, like Angle (a type of Quantity), that are
    # entries in the dict representation of a SkyCoord.
    if isinstance(value, UnitBase | Quantity):
        return str(value)

    try:
        rep = dict_rep(value)
    except AttributeError:
        # Either this is not an astropy thing, in which case just return the
        # value, or this is an astropy unit. Happily, we can already serialize
        # that.
        return value if not hasattr(value, "to_string") else value.to_string()

    result = {}

    # Recurse to handle nested astropy objects
    for k, v in rep.items():
        result[k] = serialize_astropy_type(v)

    return result


class AstropyValidator:
    """
    This class can be used to annotate pydantic fields that are an astropy type
    (e.g. SkyCoord) to add basic pydantic validation and serialization.

    Examples
    --------
    >>> from typing import Annotated
    >>> from pydantic import BaseModel
    >>> from astropy.time import Time
    >>> from astropy.coordinates import SkyCoord
    >>> from stellarphot.settings.astropy_pydantic import AstropyValidator
    >>> # Making a model with a Time field
    >>> class TimeModel(BaseModel):
    ...     time: Annotated[Time, AstropyValidator]
    >>> # The time must be either a Time object or a dictionary that can be used to
    >>> TimeModel(time=Time("2021-01-01T00:00:00"))
    TimeModel(time=<Time object: scale='utc' format='isot' value=...>)
    >>> # Making a model with a SkyCoord field
    >>> class SkyCoordModel(BaseModel):
    ...     coord: Annotated[SkyCoord, AstropyValidator]
    >>> # The coord must be either a SkyCoord object or a dictionary that can be used to
    >>> SkyCoordModel(coord=SkyCoord("00h42m44.3s +41d16m9s"))
    SkyCoordModel(coord=<SkyCoord (ICRS): (ra, dec) in deg...>)
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type,
        _handler,
    ):
        def astropy_object_from_dict(value):
            """
            This is NOT the right way to be doing this when there are nested
            definitions, e.g. in a SkyCoord where the RA and Dec are each
            an angle, which is not a native python type. For now, the serialization
            has been short-circuited compared to what serialization to a Table would
            do to get this working and released.
            """
            return source_type.info._construct_from_dict(value)

        from_dict_schema = core_schema.chain_schema(
            [
                core_schema.dict_schema(),
                core_schema.no_info_plain_validator_function(astropy_object_from_dict),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_dict_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(source_type),
                    from_dict_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                serialize_astropy_type
            ),
        )
